# This module helps with debugging Frescobaldi.
#
# Start a Python shell
# Enter: from frescobaldi.debug import *
# This runs Frescobaldi, installs some nice __repr__() methods, connects some
# signals to debug-print functions, and imports the most important modules such
# as app.



import sys

from frescobaldi.__main__ import main

from . import toplevel
toplevel.install()

import app
import document


def doc_repr(self):
    index = app.documents.index(self)
    return f'<Document #{index} "{self.url().toString()}">'
document.Document.__repr__ = doc_repr

@app.documentCreated.connect
def doc_created(doc):
    print("created:", doc)

@app.documentLoaded.connect
def doc_loaded(doc):
    print("loaded:", doc)

@app.documentClosed.connect
def doc_closed(doc):
    print("closed:", doc)

@app.jobStarted.connect
def job_started(doc, job):
    print('job started:', doc)
    print(job.command)

@app.jobFinished.connect
def job_finished(doc, job, success):
    print('job finished', doc)
    print('success:', success)


# more to add...


# delete unneeded stuff
del doc_repr, doc_created, doc_loaded, doc_closed, job_started, job_finished

def modules():
    """Print the list of loaded modules."""
    print('\n'.join(v.__name__ for k, v in sorted(sys.modules.items()) if v is not None))


# avoid builtins._ being overwritten
sys.displayhook = app.displayhook

# instantiate app and create a mainwindow, etc
main(debug=True)

# be friendly and import Qt stuff
# suppress ruff F403 lint check
from PyQt6.QtCore import *  # noqa: F403
from PyQt6.QtGui import *  # noqa: F403



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Abstract base class for a PDF viewer widget.
"""


import collections
import os
import weakref

from PyQt6.QtCore import pyqtSignal, QMargins, QPoint, QRect, Qt, QTimer, QUrl
from PyQt6.QtGui import QCursor, QTextCharFormat
from PyQt6.QtWidgets import (
    QToolTip, QVBoxLayout, QHBoxLayout, QWidget, QToolBar)

import qpageview
import qpageview.layout
import qpageview.highlight
import qpageview.rubberband
import pagedview

import app
import helpers
import textedit
import textformats
import contextmenu
import viewhighlighter
import widgets.dialog
import userguide

from . import abstractviewwidget
from . import documents
from . import pointandclick


class AbstractPdfWidget(abstractviewwidget.AbstractViewWidget):
    """Widget containing the qpageview.View."""

    # TODO: As much as possible should be moved to the base class.

    zoomChanged = pyqtSignal(int, float) # mode, scale

    def __init__(self, panel):
        """Creates the PDF View for the panel."""
        super().__init__(panel)
        self.actionCollection = panel.actionCollection
        self.createProtectedFields()
        self.createLayout()
        self._toolbar = self.createToolbar()
        self._main_layout.addWidget(self._toolbar)
        self.createHighlighters()
        self.createView()
        self.createContextMenu()
        self.connectSlots()

        # load current session when the widget is created
        import sessions
        cs = sessions.currentSession()
        if cs:
            self.slotSessionChanged(cs)
        self.readSettings()

        userguide.openWhatsThis(self)
        app.translateUI(self)

    def createProtectedFields(self):
        """Create the empty protected fields that will hold actual data."""
        self._links = None
        self._clicking_link = False
        self._toolbar = None

    def createLayout(self):
        """Set up the main layout component."""
        self._main_layout = layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def createToolbar(self):
        """Factory method to create a toolbar."""
        raise NotImplementedError()

    def createHighlighters(self):
        self._highlightFormat = QTextCharFormat()
        self._highlightMusicFormat = qpageview.highlight.Highlighter()
        self._highlightRange = None

    def createView(self):
        """Creates the actual View instance."""
        self.view = pagedview.PagedView(self)
        self.view.setRubberband(qpageview.rubberband.Rubberband())
        self._main_layout.addWidget(self.view)
        self.view.setViewMode(qpageview.FitWidth)

        self.view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.view.customContextMenuRequested.connect(self.showContextMenu)

    def createContextMenu(self):
        """Creates the context menu.
        Subclasses should override this with their concrete menu."""
        raise NotImplementedError("Concrete Viewer class {} must implement 'createContextMenu'".format(
            type(self).__name__))

    def translateUI(self):
        self.setWhatsThis(
            "<p>This viewer doesn't have a What'sThis info set.</p>")

    def connectSlots(self):
        """Connects the slots of the viewer."""
        app.settingsChanged.connect(self.readSettings)
        app.sessionChanged.connect(self.slotSessionChanged)
        app.saveSessionData.connect(self.slotSaveSessionData)

        # react if cursor of current text document moves
        self.parent().mainwindow().currentViewChanged.connect(self.slotCurrentViewChanged)
        view = self.parent().mainwindow().currentView()
        if view:
            self.slotCurrentViewChanged(view)

        self.view.linkClicked.connect(self.slotLinkClicked)
        self.view.linkHovered.connect(self.slotLinkHovered)
        self.view.linkLeft.connect(self.slotLinkLeft)
        self.view.linkHelpRequested.connect(self.slotLinkHelpRequested)

    def viewerName(self):
        """Return the viewerName() attribute of the panel."""
        if not hasattr(self, '_viewerName'):
            self._viewerName = self.parent().viewerName()
        return self._viewerName

    def sizeHint(self):
        """Returns the initial size the PDF (Music) View prefers."""
        return self.parent().mainwindow().size() / 2

    def currentViewdoc(self):
        """Returns the current Document or None."""
        return self.view.document()

    def openViewdoc(self, doc):
        """Opens a qpageview.Document instance."""
        try:
            self.view.setDocument(doc)
            doc.ispresent = True
            self._links = pointandclick.links(doc)
        except OSError:
            # the file is not found on the given path
            dlg = widgets.dialog.Dialog(buttons=('yes', 'no'))
            dlg.setWindowTitle("Missing file")
            fn = doc.filename()
            dlg.setMessage(_("The file {filename} is missing.\n\n"
                "Do you want to remove the filename from the list?").format(filename=fn))
            dlg.setIcon("question")
            dlg.setToolTip(_(
                "Answering 'No' will give you a chance to restore the "
                "file without having to re-add it."))
            if dlg.exec():
                mds = self.actionCollection.viewer_document_select
                mds.removeViewdoc(doc)
            else:
                doc.ispresent = False

    def clear(self):
        """Empties the view."""
        self._links = None
        self._highlightRange = None
        self.view.clear()

    def readSettings(self):
        """Reads the settings from the user's preferences."""
        # background and highlight colors of music view
        colors = textformats.formatData('editor').baseColors
        self._highlightMusicFormat.color = colors['musichighlight']
        color = colors['selectionbackground']
        color.setAlpha(128)
        self._highlightFormat.setBackground(color)

    def slotLinkClicked(self, ev, page, link):
        """Called when the use clicks a link.

        If the links is a textedit link, opens the document and puts the cursor there.
        Otherwise, call the helpers module to open the destination.

        """
        if ev.button() == Qt.MouseButton.RightButton:
            return
        cursor = self._links.cursor(link, True)
        if cursor:
            if ev.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                import editinplace
                editinplace.edit(self, cursor, ev.globalPos())
            else:
                import browseriface
                mainwindow = self.parent().mainwindow()
                self._clicking_link = True
                browseriface.get(mainwindow).setTextCursor(cursor, findOpenView=True)
                self._clicking_link = False
                import widgets.blink
                widgets.blink.Blinker.blink_cursor(mainwindow.currentView())
                mainwindow.activateWindow()
                mainwindow.currentView().setFocus()
        elif link.url and not link.url.startswith('textedit:'):
            helpers.openUrl(QUrl(link.url))
        elif link.targetPage != -1 and not link.isExternal:
            self.view.setCurrentPageNumber(link.targetPage)

    def slotLinkHovered(self, page, link):
        """Called when the mouse hovers a link.

        If the links points to the current editor document, the token(s) it points
        at are highlighted using a transparent selection color.

        The highlight shows for a few seconds but disappears when the mouse moves
        off the link or when the link is clicked.

        """
        cursor = self._links.cursor(link)
        if not cursor or cursor.document() != self.parent().mainwindow().currentDocument():
            return

        # highlight token(s) at this cursor
        cursors = pointandclick.positions(cursor)
        if cursors:
            view = self.parent().mainwindow().currentView()
            viewhighlighter.highlighter(view).highlight(self._highlightFormat, cursors, 2, 5000)

    def slotLinkLeft(self):
        """Called when the mouse moves off a previously highlighted link."""
        view = self.parent().mainwindow().currentView()
        viewhighlighter.highlighter(view).clear(self._highlightFormat)

    def slotLinkHelpRequested(self, ev, page, link):
        """Called when a ToolTip wants to appear above the hovered link."""
        pos = self.view.viewport().mapToGlobal(ev.pos())
        if link.url:
            cursor = self._links.cursor(link)
            if cursor:
                import documenttooltip
                text = documenttooltip.text(cursor)
            else:
                l = textedit.link(link.url)
                if l:
                    text = f"{os.path.basename(l.filename)} ({l.line}:{l.column})"
                else:
                    text = link.url
        elif link.targetPage != -1:
            text = _("Page {num}").format(num=link.targetPage)
            if link.isExternal:
                text = link.fileName + "\n" + text
        else:
            return
        QToolTip.showText(pos, text, self.view, page.linkRect(link))

    def slotSessionChanged(self, name):
        """Called whenever the current session is changed
        (also on application startup or after a session is created).
        If the session already exists load manuscripts from the
        session object and load them in the viewer."""
        if name:
            import sessions
            import qsettings
            g = sessions.sessionGroup(name)
            if g.contains("urls"): # the session is not new
                ds = self.actionCollection.viewer_document_select
                ds.removeAllViewdocs(update = False)
                self.clear()
                viewdocs = []
                files_key = f"{self.viewerName()}-documents"
                active_file = ""
                for i in range(g.beginReadArray(files_key)):
                    g.setArrayIndex(i)
                    filename = g.value("filename", "", str)
                    if filename:
                        doc = pagedview.loadPdf(filename)
                        viewdocs.append(doc)
                        # can it load?
                        doc.ispresent = os.path.isfile(filename)
                        if g.value("isactive", False, bool):
                            active_file = filename
                        props = self.view.properties().load(g)
                        self.view.documentPropertyStore.set(doc, props)
                ds.loadViewdocs(viewdocs,
                    active_viewdoc = active_file,
                    sort = False) # may be replaced by a Preference

    def slotSaveSessionData(self):
        """Saves the filenames and positions of the open manuscripts.
        If a file doesn't have a position (because it hasn't been moved or
        shown) a default position is stored."""
        import sessions
        g = sessions.currentSessionGroup()
        if g:
            # TODO: cleanup for multi-file documents later
            files_key = f"{self.viewerName()}-documents"
            docs = self.actionCollection.viewer_document_select.viewdocs()
            if docs:
                current = self.currentViewdoc()
                g.beginWriteArray(files_key, len(docs))
                for i, doc in enumerate(docs):
                    g.setArrayIndex(i)
                    g.setValue("filename", doc.filename())
                    g.setValue("isactive", doc is current)
                    if doc is current:
                        self.view.documentPropertyStore.set(doc,
                            self.view.properties().get(self.view))
                    props = self.view.documentPropertyStore.get(doc)
                    if props:
                        props.save(g)
                g.endArray()
            else:
                g.remove(files_key)

    def slotCurrentViewChanged(self, view, old=None):
        self.view.clearHighlight(self._highlightMusicFormat)
        if old:
            old.cursorPositionChanged.disconnect(self.slotCursorPositionChanged)
        view.cursorPositionChanged.connect(self.slotCursorPositionChanged)

    def slotCursorPositionChanged(self):
        """Called when the user moves the text cursor."""
        self.showCurrentLinks(
            not self._clicking_link
            and self.parent().actionCollection.viewer_sync_cursor.isChecked())

    def showCurrentLinks(self, scroll=False, msec=None):
        """Show objects at current text cursor.

        If scroll is True, also scrolls the view if needed. If msec is given,
        objects are highlighed that long. If not given, a default time is used.

        """
        if not self.isVisible() or not self._links:
            return # not visible of no PDF in the viewer

        view = self.parent().mainwindow().currentView()
        links = self._links.boundLinks(view.document())
        if not links:
            return # the PDF contains no references to the current text document

        s = links.indices(view.textCursor())
        if not s:
            if s is False:
                self.view.clearHighlight(self._highlightMusicFormat)
            return

        if not scroll and self._highlightRange == s and self.view.isHighlighting(self._highlightMusicFormat):
            return # don't redraw if same
        self._highlightRange = s

        # create the dict of areas
        layout = self.view.pageLayout()
        areas = collections.defaultdict(list)
        for dest in links.destinations()[s]:
            for pageNum, rect in dest:
                areas[layout[pageNum]].append(rect)

        if scroll:
            # compute the bounding rect
            boundingRect = QRect()
            for page, rects in areas.items():
                f = page.mapToPage(1, 1).rect
                pbound = QRect()
                for r in rects:
                    pbound |= f(r)
                boundingRect |= pbound.translated(page.pos())
            self.view.ensureVisible(boundingRect, QMargins(20, 20, 20, 20))

        if msec is None:
            msec = 5000 if s.stop - s.start > 1 else 2000 # show selections longer
        # RC: increased timer to give some time to the kinetic scrolling to complete.
        if self.view.isScrolling():
            msec += self.view.remainingScrollTime()
        self.view.highlight(areas, self._highlightMusicFormat, msec)

    def showContextMenu(self):
        """Called when the user right-clicks or presses the context menu key."""
        pos = self.view.mapToGlobal(QPoint(0, 0))
        link, cursor = None, None
        # mouse inside view?
        if self.view.mapFromGlobal(QCursor.pos()) in self.view.viewport().rect():
            pos = QCursor.pos()
            pos_in_viewport = self.view.viewport().mapFromGlobal(pos)
            page, link = self.view.linkAt(pos_in_viewport)
            if link:
                cursor = self._links.cursor(link, True)
        self._contextMenu.show(pos, link, cursor)

    def toolbar(self):
        """Returns the viewer's toolbar widget."""
        return self._toolbar



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A (Frescobaldi) document.

This contains the text the user can edit in Frescobaldi. In most cases it will
be a LilyPond source file, but other file types can be used as well.

There are two different document Classes: Document and EditorDocument.
Both provide a QTextDocument with additional metadata, but the EditorDocument
provides additional handling of signals that are hooked into the Frescobaldi
GUI environment. That means: use EditorDocument for documents open in the
editor, Document for "abstract" documents, for example to pass a generated
document to a job.lilypond.LilyPondJob without implicitly creating a tab.

"""


import os

from PyQt6.QtCore import QTimer, QUrl
from PyQt6.QtGui import QTextCursor, QTextDocument
from PyQt6.QtWidgets import QPlainTextDocumentLayout

import app
import util
import variables
import signals


class AbstractDocument(QTextDocument):
    """Base class for a Frescobaldi document. Not intended to be instantiated.

    Objects of subclasses can be passed to the functions in documentinfo
    or lilypondinfo etc. for additional meta information.

    """

    @classmethod
    def load_data(cls, url, encoding=None):
        """Class method to load document contents from an url.

        This is intended to open a document without instantiating one
        if loading the contents fails.

        This method returns the text contents of the url as decoded text,
        thus a unicode string.

        The line separator is always '\\n'.

        """
        filename = url.toLocalFile()

        # currently, we do not support non-local files
        if not filename:
            raise OSError("not a local file")
        with open(filename, 'rb') as f:
            data = f.read()
        text = util.decode(data, encoding)
        return util.universal_newlines(text)

    @classmethod
    def new_from_url(cls, url, encoding=None):
        """Create and return a new document, loaded from url.

        This is intended to open a new Document without instantiating one
        if loading the contents fails.

        """
        # Do this first, raises IOError if not found, without creating the document.
        if not url.isEmpty():
            data = cls.load_data(url, encoding)
        # If this did not raise, proceed to create a new document.
        d = cls(url, encoding)
        if not url.isEmpty():
            d.setPlainText(data)
            d.setModified(False)
        return d

    def __init__(self, url=None, encoding=None):
        """Create a new Document with url and encoding.

        Does not load the contents, you should use load() for that, or
        use the new_from_url() constructor to instantiate a new Document
        with the contents loaded.

        """
        if url is None:
            url = QUrl()
        super().__init__()
        self.setDocumentLayout(QPlainTextDocumentLayout(self))
        self._encoding = encoding
        self._url = url # avoid urlChanged on init
        self.setUrl(url)

    def load(self, url=None, encoding=None, keepUndo=False):
        """Load the specified or current url (if None was specified).

        Currently only local files are supported. An IOError is raised
        when trying to load a nonlocal URL.

        If loading succeeds and an url was specified, the url is made the
        current url (by calling setUrl() internally).

        If keepUndo is True, the loading can be undone (with Ctrl-Z).

        """
        if url is None:
            url = QUrl()
        u = url if not url.isEmpty() else self.url()
        text = self.load_data(u, encoding or self._encoding)
        if keepUndo:
            c = QTextCursor(self)
            c.select(QTextCursor.SelectionType.Document)
            c.insertText(text)
        else:
            self.setPlainText(text)
        self.setModified(False)
        if not url.isEmpty():
            self.setUrl(url)

    def _save(self, url, filename):
        with open(filename, "wb") as f:
            f.write(self.encodedText())
            f.flush()
            os.fsync(f.fileno())
        self.setModified(False)
        if not url.isEmpty():
            self.setUrl(url)

    def save(self, url=None, encoding=None):
        """Saves the document to the specified or current url.

        Currently only local files are supported. An IOError is raised
        when trying to save a nonlocal URL.

        If saving succeeds and an url was specified, the url is made the
        current url (by calling setUrl() internally).

        This method is never called directly but only from the overriding
        subclass methods that make further specific use of the modified results.

        """
        if url is None:
            url = QUrl()
        u = url if not url.isEmpty() else self.url()
        filename = u.toLocalFile()
        # currently, we do not support non-local files
        if not filename:
            raise OSError("not a local file")
        # keep the url if specified when we didn't have one, even if saving
        # would fail
        if self.url().isEmpty() and not url.isEmpty():
            self.setUrl(url)
        return url, filename

    def url(self):
        return self._url

    def setUrl(self, url):
        """ Change the url for this document. """
        if url is None:
            url = QUrl()
        old, self._url = self._url, url
        # number for nameless documents
        if self._url.isEmpty():
            nums = [0]
            nums.extend(doc._num for doc in app.documents if doc is not self)
            self._num = max(nums) + 1
        else:
            self._num = 0
        return old

    def encoding(self):
        return variables.get(self, "coding") or self._encoding

    def setEncoding(self, encoding):
        self._encoding = encoding

    def encodedText(self):
        """Return the text of the document as a bytes string encoded in the
        correct encoding.

        The line separator is '\\n' on Unix, '\\r\\n' on Windows.

        Useful to save to a file.

        """
        text = util.platform_newlines(self.toPlainText())
        return util.encode(text, self.encoding())

    def documentName(self):
        """Return a suitable name for this document.

        This is only to be used for display. If the url of the document is
        empty, something like "Untitled" or "Untitled (3)" is returned.

        """
        if self._url.isEmpty():
            if self._num == 1:
                return _("Untitled")
            else:
                return _("Untitled ({num})").format(num=self._num)
        else:
            return os.path.basename(self._url.path())


class Document(AbstractDocument):
    """A Frescobaldi document to be used anywhere except the main editor
    viewspace (also non-GUI jobs/operations)."""

    def save(self, url=None, encoding=None):
        url, filename = super().save(url, encoding)
        self._save(url, filename)


class EditorDocument(AbstractDocument):
    """A Frescobaldi document for use in the main editor view.
    Basically this is an AbstractDocument with signals added."""

    urlChanged = signals.Signal() # new url, old url
    closed = signals.Signal()
    loaded = signals.Signal()
    saving = signals.SignalContext()
    saved = signals.Signal()
    # this is like contentsChanged, but emitted only once for changes
    # occurring within a short time period, to avoid repeatedly triggering
    # slots while the user is typing
    changesStopped = signals.Signal()

    @classmethod
    def new_from_url(cls, url, encoding=None):
        d = super().new_from_url(url, encoding)
        if not url.isEmpty():
            d.loaded()
            app.documentLoaded(d)
        return d

    def __init__(self, url=None, encoding=None):
        super().__init__(url, encoding)
        self.modificationChanged.connect(self.slotModificationChanged)
        app.documents.append(self)
        app.documentCreated(self)
        # this timer is restarted after each change to the document;
        # the changesStopped signal is emitted once it times out
        self._changeTimer = QTimer(singleShot=True,
                                   interval=900, # msec
                                   timeout=self.changesStopped)
        self.contentsChanged.connect(self._changeTimer.start)

    def slotModificationChanged(self):
        app.documentModificationChanged(self)

    def close(self):
        self.closed()
        app.documentClosed(self)
        app.documents.remove(self)

    def load(self, url=None, encoding=None, keepUndo=False):
        super().load(url, encoding, keepUndo)
        self.loaded()
        app.documentLoaded(self)

    def save(self, url=None, encoding=None):
        url, filename = super().save(url, encoding)
        with self.saving(), app.documentSaving(self):
            self._save(url, filename)
        self.saved()
        app.documentSaved(self)

    def setUrl(self, url):
        old = super().setUrl(url)
        if url != old:
            self.urlChanged(url, old)
            app.documentUrlChanged(self, url, old)

    def cursorAtPosition(self, line, column=None):
        """Return a new QTextCursor set to the line and column given (each starting at 1).

        This method avoids common pitfalls associated with arbitrarily setting the cursor
        position via setCursorPosition.
        - The cursor will be set at a vaid position in a valid block.
        - Reasonable defaults are used for under/over-limit input.
        - Character counting based on UTF-8 matches LilyPond and Python conventions.
        - The cursor will not be set in the middle of a surrogate pair or composed glyph.

        """
        if line < 1:
            line = column = 1
        elif not column or column < 1:
            column = 1
        cursor = QTextCursor(self)
        block = self.findBlockByNumber(line - 1)
        if block.isValid():
            line_text = block.text()
            if len(line_text) >= column:
                qchar_offset = len(line_text[:column - 1].encode('utf_16_le')) // 2
                cursor.setPosition(block.position() + qchar_offset)
                # Escape to in front of what might be the middle of a composed glyph.
                cursor.movePosition(QTextCursor.MoveOperation.NextCharacter)
                cursor.movePosition(QTextCursor.MoveOperation.PreviousCharacter)
            else:
                cursor.setPosition(block.position())
                cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock)
        else:
            cursor.movePosition(QTextCursor.MoveOperation.End)
        return cursor

    def isChanging(self):
        """Return whether the document is currently changing.

        This returns True for a brief period after each contentsChanged
        signal is emitted. You can use this to block or delay operations
        while the user is typing.

        """
        return self._changeTimer.isActive()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Helper application stuff.
"""


import os
import platform
import re
import subprocess
import sys

from PyQt6.QtCore import QSettings
from PyQt6.QtGui import QDesktopServices
from PyQt6.QtWidgets import QMessageBox

def shell_split(cmd):
    """Split a string like the UNIX shell, returning a list.

    Double quoted parts are kept together with the quotes removed.

    """
    # remove double quotes, keeping quoted parts together
    # (because shlex does not yet work with unicode...)
    return [item.replace('"', '') for item in
        re.findall(r'[^\s"]*"[^"]*"[^\s"]*|[^\s"]+', cmd)]


def command(type):
    """Returns the command for the specified type as a list.

    Returns None if no command was specified.

    """
    cmd = QSettings().value("helper_applications/" + type, "")
    if not cmd:
        return

    if os.path.isabs(cmd) and os.access(cmd, os.X_OK):
        return [cmd]

    return shell_split(cmd)


def openUrl(url, type="browser"):
    """Opens the specified QUrl, using the specified type."""
    # be sure to pick a suitable type
    if url.scheme() == "mailto":
        type = "email"
    elif type == "browser" and url.toLocalFile():
        ext = os.path.splitext(url.toLocalFile())[1].lower()
        if os.path.isdir(url.toLocalFile()):
            type = "directory"
        elif ext == '.pdf':
            type = "pdf"
        elif ext in ('.png', '.jpg', '.jpeg'):
            type = "image"
        elif ext in ('.midi', '.mid'):
            type = "midi"

    # get the command
    cmd = command(type)
    if not cmd:
        if type != "shell":
            QDesktopServices.openUrl(url)
            return
        for cmd in terminalCommands():
            break # get the first

    prog = cmd.pop(0)

    workdir = None
    if url.toLocalFile():
        workdir = url.toLocalFile()
        if type != "shell":
            workdir = os.path.dirname(workdir)

    if any('$f' in a or '$u' in a for a in cmd):
        cmd = [a.replace('$u', url.toString())
                if '$u' in a else a.replace('$f', url.toLocalFile())
                for a in cmd]
    elif type in ("browser", "email"):
        cmd.append(url.toString())
    elif type != "shell":
        cmd.append(url.toLocalFile())

    try:
        subprocess.Popen([prog] + cmd, cwd=workdir)
    except OSError:
        QMessageBox.critical(None, _("Error"), _(
            "Could not start {program}.\n"
            "Please check path and permissions.").format(program=prog))


def terminalCommands():
    """Yields suitable commands to open a terminal/shell window.

    There is always yielded at least one, which is suitable as default.

    """
    if platform.system() == "Windows":
        yield ['cmd.exe']
    elif platform.system() == "Darwin":
        yield ['open', '-a', 'Terminal', '$f']
    else:
        # find a default linux terminal
        paths = os.environ.get('PATH', os.defpath).split(os.pathsep)
        for cmd in (
            ['lxterminal', '--working-directory=$f'],
            ['xfce4-terminal', '--working-directory=$f'],
            ['konsole', '--workdir', '$f'],
            ['gnome-terminal', '--working-directory=$f'],
            ):
            for p in paths:
                if p:
                    prog = os.path.join(p, cmd[0])
                    if os.access(prog, os.X_OK):
                        yield cmd
        yield ['xterm']



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Internationalization of Frescobaldi.
"""

import builtins
import gettext
from pathlib import Path

modir = __path__[0]

class UnknownLanguageError(Exception):
    """Raised when calling translator() on an unknown language."""
    pass

def available():
    """Returns a list of language shortnames for which a MO file is available.

    Note that this is not the full list of available languages; there
    are also English ("en") and the special language "C".
    """
    return [str(path.parent.stem) for path in Path(modir).rglob("LC_MESSAGES")]

def translator(language, domain="frescobaldi"):
    """Returns a function that can translate messages using the specified language.

    The value "C" or "en" or "en_XXX" causes a dummy translator that returns English
    strings untranslated to be returned. Otherwise, the language must have a translation,
    meaning that either `language` or `language.split('_')[0]` is in the list
    returned by `available()`.

    The returned function can be called with one to four arguments:

    - message
    - context, message
    - message, plural_message, count
    - context, message, plural_message, count

    In all cases a single string (the translation) is returned.
    """
    if language == "C" or language == "en" or language.split('_')[0] == "en":
        catalog = gettext.NullTranslations()
    elif language in available() or language.split('_')[0] in available():
        catalog = gettext.translation(domain, localedir=modir,
                                      languages=[language])
    else:
        raise UnknownLanguageError
    funcs = (None, catalog.gettext, catalog.pgettext, catalog.ngettext, catalog.npgettext)
    return lambda *args: funcs[len(args)](*args)

def install(language):
    """Installs the translations for the given language."""
    builtins._ = translator(language)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The score settings widget.
"""


from PyQt6.QtCore import QSettings, Qt
from PyQt6.QtWidgets import (QCheckBox, QComboBox, QGridLayout, QGroupBox,
                             QHBoxLayout, QLabel, QVBoxLayout, QWidget,
                             QButtonGroup, QRadioButton)

import app
import i18n.setup
import language_names
import listmodel
import lilypondinfo

from . import scoreproperties

class SettingsWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        grid = QGridLayout()
        self.setLayout(grid)

        self.scoreProperties = ScoreProperties(self)
        self.generalPreferences = GeneralPreferences(self)
        self.lilyPondPreferences = LilyPondPreferences(self)
        self.instrumentNames = InstrumentNames(self)
        self.midiOutput = MidiOutput(self)

        grid.addWidget(self.scoreProperties, 0, 0)
        grid.addWidget(self.generalPreferences, 0, 1)
        grid.addWidget(self.lilyPondPreferences, 1, 0, 2, 1)
        grid.addWidget(self.instrumentNames, 1, 1)
        grid.addWidget(self.midiOutput, 2, 1)

    def clear(self):
        self.scoreProperties.tempo.clear()
        self.scoreProperties.keyNote.setCurrentIndex(0)
        self.scoreProperties.keyMode.setCurrentIndex(0)
        self.scoreProperties.pickup.setCurrentIndex(0)


class ScoreProperties(QGroupBox, scoreproperties.ScoreProperties):
    def __init__(self, parent):
        super().__init__(parent)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.createWidgets()
        self.layoutWidgets(layout)

        app.translateUI(self)

        scorewiz = self.window()
        scorewiz.pitchLanguageChanged.connect(self.setPitchLanguage)
        self.setPitchLanguage(scorewiz.pitchLanguage())

        self.loadSettings()
        self.window().finished.connect(self.saveSettings)

    def translateUI(self):
        self.translateWidgets()
        self.setTitle(_("Score properties"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/scoreproperties')
        self.metronomeRound.setChecked(s.value('round_metronome', True, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/scoreproperties')
        s.setValue('round_metronome', self.metronomeRound.isChecked())


class GeneralPreferences(QGroupBox):
    def __init__(self, parent):
        super().__init__(parent)

        layout = QGridLayout()
        self.setLayout(layout)

        self.typq = QCheckBox()
        self.relpitch = QCheckBox()
        self.tagl = QCheckBox()
        self.barnum = QCheckBox()
        self.neutdir = QCheckBox()
        self.metro = QCheckBox()

        # paper size
        self.paperSizeLabel = QLabel()
        self.paper = QComboBox()
        self.paper.addItems(paperSizes)
        self.paper.activated.connect(self.slotPaperChanged)

        # paper orientation
        self.paperOrientationLabel = QLabel()
        self.paperOrientation = QComboBox()
        self.paperOrientation.addItems([''] * 3)

        layout.addWidget(self.typq, 0, 0, 1, 2)
        layout.addWidget(self.relpitch, 1, 0, 1, 2)
        layout.addWidget(self.tagl, 2, 0, 1, 2)
        layout.addWidget(self.barnum, 3, 0, 1, 2)
        layout.addWidget(self.neutdir, 4, 0, 1, 2)
        layout.addWidget(self.metro, 5, 0, 1, 2)
        layout.addWidget(self.paperSizeLabel, 6, 0)
        layout.addWidget(self.paper, 6, 1)
        layout.addWidget(self.paperOrientationLabel, 7, 0)
        layout.addWidget(self.paperOrientation, 7, 1)

        app.translateUI(self)

        self.loadSettings()
        self.window().finished.connect(self.saveSettings)

    def translateUI(self):
        self.setTitle(_("General preferences"))
        self.typq.setText(_("Use typographical quotes"))
        self.typq.setToolTip(_(
            "Replace normal quotes in titles with nice typographical quotes."))
        self.relpitch.setText(_("Use \\relative with pitch"))
        self.relpitch.setToolTip(_(
            "Write a default pitch after the \\relative command."))
        self.tagl.setText(_("Remove default tagline"))
        self.tagl.setToolTip(_(
            "Suppress the default tagline output by LilyPond."))
        self.barnum.setText(_("Remove bar numbers"))
        self.barnum.setToolTip(_(
            "Suppress the display of measure numbers at the beginning of "
            "every system."))
        self.neutdir.setText(_("Smart neutral stem direction"))
        self.neutdir.setToolTip(_(
            "Use a logical direction (up or down) for stems on the middle "
            "line of a staff."))
        self.metro.setText(_("Show metronome mark"))
        self.metro.setToolTip(_(
            "If checked, show the metronome mark at the beginning of the "
            "score. The MIDI output also uses the metronome setting."))
        # paper size and orientation:
        self.paperSizeLabel.setText(_("Paper size:"))
        self.paper.setItemText(0, _("Default"))
        self.paperOrientationLabel.setText(_("Orientation:"))
        self.paperOrientation.setItemText(0, _("Regular"))
        self.paperOrientation.setItemText(1, _("Landscape"))
        self.paperOrientation.setItemText(2, _("Rotated"))
        self.paperOrientation.setItemData(0, _("Regular portrait orientation"), Qt.WindowType.ToolTip)
        self.paperOrientation.setItemData(1, _(
            "Set paper orientation to landscape while keeping upright printing orientation."),
            Qt.WindowType.ToolTip)
        self.paperOrientation.setItemData(2, _("Rotate print on regular paper."), Qt.WindowType.ToolTip)

    def slotPaperChanged(self, index):
        self.paperOrientation.setEnabled(bool(index))

    def getPaperSize(self):
        """Returns the configured papersize or the empty string for default."""
        return paperSizes[self.paper.currentIndex()]

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/preferences')
        self.typq.setChecked(s.value('typographical_quotes', True, bool))
        self.relpitch.setChecked(s.value('relative_pitch', True, bool))
        self.tagl.setChecked(s.value('remove_tagline', False, bool))
        self.barnum.setChecked(s.value('remove_barnumbers', False, bool))
        self.neutdir.setChecked(s.value('smart_neutral_direction', False, bool))
        self.metro.setChecked(s.value('metronome_mark', False, bool))
        psize = s.value('paper_size', '', str)
        enable = bool(psize and psize in paperSizes)
        self.paper.setCurrentIndex(paperSizes.index(psize) if enable else 0)
        orientation = s.value('paper_rotation', 0, int)
        self.paperOrientation.setCurrentIndex(orientation)
        self.paperOrientation.setEnabled(enable)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/preferences')
        s.setValue('typographical_quotes', self.typq.isChecked())
        s.setValue('relative_pitch', self.relpitch.isChecked())
        s.setValue('remove_tagline', self.tagl.isChecked())
        s.setValue('remove_barnumbers', self.barnum.isChecked())
        s.setValue('smart_neutral_direction', self.neutdir.isChecked())
        s.setValue('metronome_mark', self.metro.isChecked())
        s.setValue('paper_size', paperSizes[self.paper.currentIndex()])
        s.setValue('paper_rotation', self.paperOrientation.currentIndex())


class MidiOutput(QGroupBox):
    def __init__(self, parent):
        super().__init__(parent, checkable=True, checked=True)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.separateScore = QCheckBox()
        layout.addWidget(self.separateScore)

        app.translateUI(self)

        self.loadSettings()
        self.window().finished.connect(self.saveSettings)

    def translateUI(self):
        self.setTitle(_("Create MIDI output"))
        self.setToolTip(_(
            "Create a MIDI file in addition to the PDF file."))

        self.separateScore.setText(_("Place in a separate \\score block"))
        self.separateScore.setToolTip(_(
            "Create a separate \\score block for MIDI output."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/preferences')    # for backwards compatibility
        self.setChecked(s.value('midi', True, bool))
        self.separateScore.setChecked(s.value('separateMidi', False, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/preferences')    # for backwards compatibility
        s.setValue('midi', self.isChecked())
        s.setValue('separateMidi', self.separateScore.isChecked())


class InstrumentNames(QGroupBox):
    def __init__(self, parent):
        super().__init__(parent, checkable=True, checked=True)

        grid = QGridLayout()
        self.setLayout(grid)

        self.firstSystemLabel = QLabel()
        self.firstSystem = QComboBox()
        self.firstSystemLabel.setBuddy(self.firstSystem)
        self.otherSystemsLabel = QLabel()
        self.otherSystems = QComboBox()
        self.otherSystemsLabel.setBuddy(self.otherSystems)
        self.languageLabel = QLabel()
        self.language = QComboBox()
        self.languageLabel.setBuddy(self.language)

        self.firstSystem.setModel(listmodel.ListModel(
            (lambda: _("Long"), lambda: _("Short"), lambda: _("None")), self.firstSystem,
            display = listmodel.translate))
        self.otherSystems.setModel(listmodel.ListModel(
            (lambda: _("Long"), lambda: _("Short"), lambda: _("None")), self.otherSystems,
            display = listmodel.translate))

        self._langs = l = ['','C']
        l.extend(sorted(i18n.available()))
        def display(lang):
            if lang == 'C':
                return _("English (untranslated)")
            elif not lang:
                return _("Default")
            return language_names.languageName(lang, i18n.setup.current())
        self.language.setModel(listmodel.ListModel(l, self.language, display=display))

        grid.addWidget(self.firstSystemLabel, 0, 0)
        grid.addWidget(self.firstSystem, 0, 1)
        grid.addWidget(self.otherSystemsLabel, 1, 0)
        grid.addWidget(self.otherSystems, 1, 1)
        grid.addWidget(self.languageLabel, 2, 0)
        grid.addWidget(self.language, 2, 1)
        app.translateUI(self)
        self.loadSettings()
        self.window().finished.connect(self.saveSettings)

    def translateUI(self):
        self.setTitle(_("Instrument names"))
        self.firstSystemLabel.setText(_("First system:"))
        self.otherSystemsLabel.setText(_("Other systems:"))
        self.languageLabel.setText(_("Language:"))
        self.firstSystem.setToolTip(_(
            "Use long or short instrument names before the first system."))
        self.otherSystems.setToolTip(_(
            "Use short, long or no instrument names before the next systems."))
        self.language.setToolTip(_(
            "Which language to use for the instrument names."))
        self.firstSystem.model().update()
        self.otherSystems.model().update()
        self.language.model().update()

    def getLanguage(self):
        """Returns the language the user has set.

        '' means:  default (use same translation as system)
        'C' means: English (untranslated)
        or a language code that is available in Frescobaldi's translation.

        """
        return self._langs[self.language.currentIndex()]

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/instrumentnames')
        self.setChecked(s.value('enabled', True, bool))
        allow = ['long', 'short', 'none']
        first = s.value('first', '', str)
        self.firstSystem.setCurrentIndex(allow.index(first) if first in allow else 0)
        other = s.value('other', '', str)
        self.otherSystems.setCurrentIndex(allow.index(other) if other in allow else 2)
        language = s.value('language', '', str)
        self.language.setCurrentIndex(self._langs.index(language) if language in self._langs else 0)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('scorewiz/instrumentnames')
        s.setValue('enable', self.isChecked())
        s.setValue('first', ('long', 'short', 'none')[self.firstSystem.currentIndex()])
        s.setValue('other', ('long', 'short', 'none')[self.otherSystems.currentIndex()])
        s.setValue('language', self._langs[self.language.currentIndex()])


class LilyPondPreferences(QGroupBox):
    def __init__(self, parent):
        super().__init__(parent)

        grid = QGridLayout()
        self.setLayout(grid)

        self.pitchLanguageLabel = QLabel()
        self.pitchLanguage = QComboBox()
        self.versionLabel = QLabel()
        self.version = QComboBox(editable=True)

        self.pitchLanguage.addItem('')
        self.pitchLanguage.addItems([lang.title() for lang in sorted(scoreproperties.keyNames)])
        self.version.addItem(lilypondinfo.preferred().versionString())
        for v in lilypondVersions:
            if v != lilypondinfo.preferred().versionString():
                self.version.addItem(v)

        grid.addWidget(self.pitchLanguageLabel, 0, 0)
        grid.addWidget(self.pitchLanguage, 0, 1)
        grid.addWidget(self.versionLabel, 1, 0)
        grid.addWidget(self.version, 1, 1)

        self.pitchLanguage.activated.connect(self.slotPitchLanguageChanged)
        app.translateUI(self)
        self.loadSettings()
        self.window().finished.connect(self.saveSettings)

    def translateUI(self):
        self.setTitle(_("LilyPond"))
        self.pitchLanguageLabel.setText(_("Pitch name language:"))
        self.pitchLanguage.setToolTip(_(
            "The LilyPond language you want to use for the pitch names."))
        self.pitchLanguage.setItemText(0, _("Default"))
        self.versionLabel.setText(_("Version:"))
        self.version.setToolTip(_(
            "The LilyPond version you will be using for this document."))

    def slotPitchLanguageChanged(self, index):
        if index == 0:
            language = ''
        else:
            language = self.pitchLanguage.currentText().lower()
        self.window().setPitchLanguage(language)

    def loadSettings(self):
        language = self.window().pitchLanguage()
        languages = list(sorted(scoreproperties.keyNames))
        index = languages.index(language) + 1 if language in languages else 0
        self.pitchLanguage.setCurrentIndex(index)

    def saveSettings(self):
        QSettings().setValue('scorewiz/lilypond/pitch_language', self.window().pitchLanguage())


lilypondVersions = ["2.26.0", "2.24.0", "2.22.0", "2.20.0", "2.18.0", "2.16.0", "2.14.0", "2.12.0"]
paperSizes = ['', 'a3', 'a4', 'a5', 'a6', 'a7', 'legal', 'letter', '11x17']



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Animated "waiting" overlay to be attached to arbitrary widgets.

Create an Overlay object and make sure it has a QWidget as parent.
With start() and stop() a (semi-) transparent waiting animation is
centered over the parent widget.
"""

# Modified after
# https://wiki.python.org/moin/PyQt/A%20full%20widget%20waiting%20indicator

from PyQt6.QtCore import (
    Qt,
)
from PyQt6.QtGui import (
    QBrush,
    QColor,
    QPainter,
    QPen
)
from PyQt6.QtWidgets import (
    QWidget
)

class Overlay(QWidget):
    """
    Transparent overlay widget displaying a "waiting" animation
    centered over its parent widget.
    - interval: update interval in ms
    - transparency: transparency (0-255) of the effect background.
    """

    # TODO: Add more configuration options (background and foreground colors)
    def __init__(
        self,
        parent=None,
        interval=75,
        transparency=200
    ):
        super().__init__(parent)
        self._counter = 0
        self._timerId = None
        self.setTransparency(transparency)
        self.setInterval(interval)
        self.hide()

    def paintEvent(self, event):
        """Repaint the widget.

        First create a (semi-) transparent background for the whole
        parent widget, then paint six small circles, with one
        colored differently.
        """

        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.fillRect(
            event.rect(), QColor(255, 255, 255, self._transparency)
        )
        painter.setPen(QPen(Qt.PenStyle.NoPen))
        painter.translate(self.rect().center())

        for i in range(6):
            if (self._counter // 5) % 6 == i:
                painter.setBrush(
                    QBrush(QColor(127 + (self._counter % 5)*32, 127, 127))
                )
            else:
                painter.setBrush(QBrush(QColor(127, 127, 127)))
            painter.drawEllipse(30, -10, 20, 20)
            painter.rotate(60)
        painter.end()

    def setInterval(self, value):
        """Change the timer interval (ms)."""
        self._interval = value

    def setTransparency(self, value):
        """Set the transparency of the animation background,
        input is limited to 0-255."""
        value = max(0, min(255, value))
        self._transparency = 255 - value

    def start(self):
        """Show the overlay and start the animation."""
        self._timerId = self.startTimer(self._interval)
        self.resize(self.parent().size())
        self.show()

    def stop(self):
        """Stop the animation and hide the overlay."""
        if self._timerId is not None:
            self.killTimer(self._timerId)
            self._timerId = None
        self.hide()
        self._counter = 0

    def timerEvent(self, ev):
        self._counter += 1
        self.resize(self.parent().size())
        self.update()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A LilyPondJob runs LilyPond and captures the output
to get it later or to have a log follow it.
"""

import codecs
import glob
import os
import shutil
import sys

from PyQt6.QtCore import QSettings, QUrl

import ly.document
import ly.docinfo

import document
import documentinfo
from . import Job
import lilypondinfo
import util


def parse_d_option(token):
    """Parse a -d option into its bare name and a Boolean value. For example
    '-dno-point-and-click' will be parsed into 'point-and-click' and False
    """
    token = token[2:]
    value = True
    if token.startswith('no-'):
        token = token[3:]
        value = False
    return token, value


def serialize_d_options(d_options, ordered=False):
    """Compose an (un)ordered list of -d options in their actual
    command line syntax."""
    result = []
    keys = sorted(d_options.keys()) if ordered else d_options.keys()
    for key in keys:
        result.append('-d{}{}'.format(
            '' if d_options[key] else 'no-', key))
    return result


class LilyPondJob(Job):
    """Manages a LilyPond process.

    In addition to the generic functionality of the job.Job class
    this class performs basic LilyPond-related configuration.

    The job always works on a *LilyPond* document, which is passed to
    LilyPond as the last command line argument. This is different from
    the way the generic Job class works.

    LilyPond Preferences are used as default values but can be overwritten
    before calling start().

    Code using the LilyPondJob class or its descendants doesn't explicitly
    deal with the 'command' list. Instead properties are set and options
    added from which the command line is implicitly composed in
    configure_command().

    """

    def __init__(self, doc, args=None, title=""):
        """Create a LilyPond job by first retrieving some context
        from the document and feeding this into job.Job's __init__()."""
        if isinstance(doc, QUrl):
            doc = document.Document(doc)
        self.document = doc
        self.document_info = docinfo = documentinfo.info(doc)
        self.lilypond_info = docinfo.lilypondinfo()
        self._d_options = {}
        self._backend_args = []
        input, self.includepath = docinfo.jobinfo(True)
        directory = os.path.dirname(input)

        super().__init__(
                encoding='utf-8',
                args=args,
                input=input,
                decode_errors='replace',
                directory=directory,
                environment=(
                    {'LD_LIBRARY_PATH': libdir}
                    if (libdir := self.lilypond_info.libdir())
                    else {}),
                title=title,
                priority=2)

        # Set default values from Preferences
        s = QSettings()
        s.beginGroup("lilypond_settings")
        self.set_d_option('delete-intermediate-files',
            s.value("delete_intermediate_files", True, bool))
        self.default_output_target = s.value(
            "default_output_target", "pdf", str)
        self.embed_source_code = s.value("embed_source_code", False, bool)
        if s.value("no_translation", False, bool):
            self.environment['LC_MESSAGES'] = 'C'
        self.set_title("{} {} [{}]".format(
            os.path.basename(self.lilypond_info.command),
            self.lilypond_info.versionString(), doc.documentName()))

    def add_additional_arg(self, arg):
        """Append an additional command line argument if it is not
        present already."""
        if arg not in self._additional_args:
            self._additional_args.append(arg)

    def add_include_path(self, path):
        """Manually add an include path to the current job."""
        self.includepath.append(path)

    def additional_args(self):
        """Additional (custom) arguments, will be inserted between
        the -d options and the include paths. May for example stem
        from the manual part of the Engrave Custom dialog."""
        return self._additional_args

    def backend_args(self):
        """Determine the target/backend type and produce appropriate args."""
        result = self._backend_args
        if not result:
            # no specific backend selected
            if self.default_output_target == "svg":
                # engrave to SVG
                result.append('-dbackend=svg')
            else:
                # engrave to PDF
                if not self.arguments():
                    # publish mode
                    if self.embed_source_code and self.lilypond_info.version() >= (2, 19, 39):
                        result.append('-dembed-source-code')
                result.append('--pdf')
        return result

    def configure_command(self):
        """Compose the command line for a LilyPond job using all options.
        Individual steps may be overridden in subclasses."""
        self.command = cmd = (
            [self.lilypond_info.abscommand() or self.lilypond_info.command])
        cmd.extend(serialize_d_options(self._d_options))
        cmd.extend(self.arguments())
        cmd.extend(self.paths(self.includepath))
        cmd.extend(self.backend_args())
        self.set_input_file()

    def d_option(self, key):
        return self._d_options.get(key, None)

    def paths(self, includepath):
        """Ensure paths have trailing slashes for Windows compatibility."""
        result = []
        for path in includepath:
            result.append('-I' + path.rstrip('/') + '/')
        return result

    def set_backend_args(self, args):
        self._backend_args = args

    def set_d_option(self, key, value=True):
        self._d_options[key] = value


class PreviewJob(LilyPondJob):
    """Represents a LilyPond Job in Preview mode."""

    def __init__(self, document, args=None, title=""):
        super().__init__(document, args, title)
        self.set_d_option('point-and-click', True)


class PublishJob(LilyPondJob):
    """Represents a LilyPond Job in Publish mode."""

    def __init__(self, document, args=None, title=""):
        super().__init__(document, args, title)
        self.set_d_option('point-and-click', False)


class LayoutControlJob(LilyPondJob):
    """Represents a LilyPond Job in Layout Control mode."""

    def __init__(self, document, args=None, title=""):
        super().__init__(document, args, title)
        # So far no further code is necessary


class VolatileTextJob(PublishJob):
    """Represents a 'volatile' LilyPond Job where the document
    is only passed in as a string. Internally a document is created
    in a temporary file, and options set to not use point-and-click.
    base_dir can be used to add a 'virtual' document Directory
    in order to use relative includes.
    """
    def __init__(self, text, title=None, base_dir=None):
        # Create temporary (document.Document object and file)
        self.directory = util.tempdir()
        filename = os.path.join(self.directory, 'document.ly')
        with open(filename, 'wb') as f:
            f.write(text.encode('utf-8'))
        url = QUrl(filename)
        url.setScheme('file')
        super().__init__(url, title=title)

        if title:
            self.set_title(title)
        if base_dir:
            self.add_include_path(base_dir)

    def resultfiles(self):
        """Returns a list of resulting file(s)"""
        #TODO: Support non-PDF compilation modes
        return glob.glob(os.path.join(self.directory, '*.pdf'))

    def cleanup(self):
        shutil.rmtree(self.directory, ignore_errors=True)


class CachedPreviewJob(PublishJob):
    """Represents a cached example LilyPond Job where the document
    is only passed in as a string. Internally a document is created
    in a cached file, and options set to not use point-and-click.
    The filename is generated as the md5 hash of the passed text, and
    the compilation is only started if a corresponding file has not
    been compiled yet.
    If a target_dir is given it is used (for example to allow persistent
    caching), otherwise an autogenerated temporary directory is used
    (and deleted upon program termination).
    base_dir can be used to add a 'virtual' document Directory
    in order to use relative includes from the 'current document'.
    """

    _target_dir = util.tempdir()

    def __init__(
        self, text, target_dir=None, title=None, base_dir=None
    ):
        import hashlib
        md = hashlib.md5()
        md.update(text.encode('utf-8'))
        self.hash_name = md.hexdigest()
        self.base_name = self.hash_name + '.ly'
        self.target_dir = target_dir or self._target_dir
        filename = os.path.join(self.target_dir, self.base_name)
        if os.path.exists(os.path.join(
            self.target_dir,
            self.hash_name + '.pdf')
        ):
            self._needs_compilation = False
        else:
            with open(filename, 'wb') as f:
                f.write(text.encode('utf-8'))
            self._needs_compilation = True
        url = QUrl(filename)
        url.setScheme('file')
        super().__init__(url, title=title)
        self.done.connect(self.remove_intermediate)

        if title:
            self.set_title(title)
        if base_dir:
            self.add_include_path(base_dir)

    def cleanup(self):
        """Do *not* remove the generated files."""
        pass

    def needs_compilation(self):
        return self._needs_compilation

    def remove_intermediate(self):
        """Remove all files from the compilation except
        the (main) .pdf and the .ly files."""
        dir = self.target_dir
        files = os.listdir(dir)
        hash_name = self.hash_name
        keep = [hash_name + '.ly', hash_name + '.pdf']
        for f in files:
            if f.startswith(hash_name) and f not in keep:
                os.remove(os.path.join(dir, f))

    def resultfiles(self):
        """
        Returns a single result file, wrapped in a list.
        This for example prevents system-wise files from
        lilypond-book-preamble to clutter the preview.
        """
        #TODO: Support non-PDF compilation modes
        output_name, _ = os.path.splitext(self.base_name)
        resultfile = os.path.join(self.directory(), output_name + '.pdf')
        if os.path.exists(resultfile):
            return [resultfile]
        else:
            return []

    def start(self):
        """Override the Job start, using cached PDF if possible."""
        if self.needs_compilation():
            super().start()
        else:
            self.done("cached")



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Functions that have to do with macOS-specific behaviour.

Importing this package does nothing; the macos.setup module sets up
the macOS-specific behaviour, such as the global menu and the file open event
handler, etc.

"""


import os
import sys


def inside_app_bundle():
    """Return True if we are inside a .app bundle."""
    # Testing for sys.frozen == 'macosx_app' (i.e. are we inside a .app bundle
    # packaged with py2app?) should be sufficient, but let's also check whether
    # we are inside a .app-bundle-like folder structure.
    return (getattr(sys, 'frozen', '') == 'macosx_app' or
        '.app/Contents/MacOS' in os.path.abspath(sys.argv[0]))

def inside_lightweight_app_bundle():
    """Return True if we are inside a lightweight .app bundle."""
    # A lightweight .app bundle (created with macos/mac-app.py without
    # the standalone option) contains a symlink to the Python interpreter
    # instead of a copy.
    return (inside_app_bundle()
            and os.path.islink(os.getcwd() + '/../MacOS/python'))

def use_osx_menu_roles():
    """Return True if macOS-specific menu roles are to be used."""
    global _use_roles
    try:
        return _use_roles
    except NameError:
        _use_roles = inside_app_bundle()
    return _use_roles



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Widget for search and replace.
"""


import bisect
import re
import weakref

from PyQt6.QtCore import QEvent, Qt
from PyQt6.QtGui import QAction, QKeySequence, QPalette, QTextCursor
from PyQt6.QtWidgets import (
    QApplication, QCheckBox, QGridLayout, QLabel, QLineEdit,
    QPushButton, QStyle, QToolButton, QWidget)

import app
import icons
import qutil
import plugin
import userguide
import cursortools
import textformats
import wordboundary
import viewhighlighter
import gadgets.borderlayout


class Search(plugin.MainWindowPlugin, QWidget):
    def __init__(self, mainwindow):
        QWidget.__init__(self, mainwindow)
        self._currentView = None
        self._positions = []
        self._positionsDirty = True
        self._replace = False  # are we in replace mode?
        self._going = False    # are we moving the text cursor?

        mainwindow.currentViewChanged.connect(self.viewChanged)
        mainwindow.actionCollection.edit_find_next.triggered.connect(self.findNext)
        mainwindow.actionCollection.edit_find_previous.triggered.connect(self.findPrevious)

        # don't inherit looks from view
        self.setFont(QApplication.font())
        self.setPalette(QApplication.palette())

        grid = QGridLayout(spacing=2)
        grid.setContentsMargins(4, 0, 4, 0)
        grid.setVerticalSpacing(0)
        self.setLayout(grid)

        self.searchEntry = QLineEdit(textChanged=self.slotSearchChanged)
        self.searchLabel = QLabel()
        self.prevButton = QToolButton(autoRaise=True, focusPolicy=Qt.FocusPolicy.NoFocus, clicked=self.findPrevious)
        self.prevButton.setIcon(icons.get('go-previous'))
        self.nextButton = QToolButton(autoRaise=True, focusPolicy=Qt.FocusPolicy.NoFocus, clicked=self.findNext)
        self.nextButton.setIcon(icons.get('go-next'))
        self.caseCheck = QCheckBox(checked=True, focusPolicy=Qt.FocusPolicy.NoFocus)
        self.regexCheck = QCheckBox(focusPolicy=Qt.FocusPolicy.NoFocus)
        self.countLabel = QLabel(alignment=Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.countLabel.setMinimumWidth(self.countLabel.fontMetrics().horizontalAdvance("9999"))
        self.closeButton = QToolButton(autoRaise=True, focusPolicy=Qt.FocusPolicy.NoFocus)
        self.hideAction = QAction(self, triggered=self.slotHide)
        self.hideAction.setShortcut(QKeySequence(Qt.Key.Key_Escape))
        self.hideAction.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogCloseButton))
        self.closeButton.setDefaultAction(self.hideAction)

        grid.addWidget(self.searchLabel, 0, 0)
        grid.addWidget(self.searchEntry, 0, 1)
        grid.addWidget(self.prevButton, 0, 2)
        grid.addWidget(self.nextButton, 0, 3)
        grid.addWidget(self.caseCheck, 0, 4)
        grid.addWidget(self.regexCheck, 0, 5)
        grid.addWidget(self.countLabel, 0, 6)
        grid.addWidget(self.closeButton, 0, 7)

        self.caseCheck.toggled.connect(self.slotSearchChanged)
        self.regexCheck.toggled.connect(self.slotSearchChanged)

        self.replaceEntry = QLineEdit()
        self.replaceLabel = QLabel()
        self.replaceButton = QPushButton(clicked=self.slotReplace)
        self.replaceAllButton = QPushButton(clicked=self.slotReplaceAll)

        grid.addWidget(self.replaceLabel, 1, 0)
        grid.addWidget(self.replaceEntry, 1, 1)
        grid.addWidget(self.replaceButton, 1, 4)
        grid.addWidget(self.replaceAllButton, 1, 5)

        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
        app.translateUI(self)

    def translateUI(self):
        self.searchLabel.setText(_("Search:"))
        self.prevButton.setToolTip(_("Find Previous"))
        self.nextButton.setToolTip(_("Find Next"))
        self.caseCheck.setText(_("&Case"))
        self.caseCheck.setToolTip(_("Case Sensitive"))
        self.regexCheck.setText(_("&Regex"))
        self.regexCheck.setToolTip(_("Regular Expression"))
        self.countLabel.setToolTip(_("The total number of matches"))
        self.hideAction.setToolTip(_("Close"))
        self.replaceLabel.setText(_("Replace:"))
        self.replaceButton.setText(_("Re&place"))
        self.replaceButton.setToolTip(_("Replaces the next occurrence of the search term."))
        self.replaceAllButton.setText(_("&All"))
        self.replaceAllButton.setToolTip(_("Replaces all occurrences of the search term in the document or selection."))

    def readSettings(self):
        data = textformats.formatData('editor')
        self.searchEntry.setFont(data.font)
        self.replaceEntry.setFont(data.font)
        p = data.palette()
        self.searchEntry.setPalette(p)
        self.replaceEntry.setPalette(p)

    def currentView(self):
        """Return the currently active View."""
        return self._currentView and self._currentView()

    def setCurrentView(self, view):
        """Set the currently active View, called by showWidget()."""
        cur = self.currentView()
        if cur:
            cur.selectionChanged.disconnect(self.slotSelectionChanged)
            cur.document().contentsChanged.disconnect(self.slotDocumentContentsChanged)
        if view:
            view.selectionChanged.connect(self.slotSelectionChanged)
            view.document().contentsChanged.connect(self.slotDocumentContentsChanged)
        self._currentView = weakref.ref(view) if view else None

    def showWidget(self):
        """Show the search widget and connect with the active View."""
        if self.isVisible():
            self.hideWidget()
        view = self.window().currentView()
        self.setCurrentView(view)
        layout = gadgets.borderlayout.BorderLayout.get(view)
        layout.addWidget(self, gadgets.borderlayout.BOTTOM)
        self.show()

    def hideWidget(self):
        """Hide the widget, but remain connected with the active View."""
        view = self.currentView()
        if view:
            self.highlightingOff()
            self.hide()
            layout = gadgets.borderlayout.BorderLayout.get(view)
            layout.removeWidget(self)

    def viewChanged(self, new):
        """Called when the user switches to another View."""
        self.setParent(None)
        self.hideWidget()
        self.setCurrentView(new)
        self.markPositionsDirty()

    def slotSelectionChanged(self):
        """Called when the user changes the selection."""
        if not self._going:
            self.markPositionsDirty()
            if self.isVisible():
                self.updatePositions()
                self.highlightingOn()

    def slotDocumentContentsChanged(self):
        """Called when the current document changes."""
        self.markPositionsDirty()
        if self.isVisible():
            self.updatePositions()
            self.highlightingOn()

    def slotHide(self):
        """Called when the close button is clicked."""
        view = self.currentView()
        if view:
            self.hideWidget()
            view.setFocus()

    def find(self):
        """Called by the main menu Find... command."""
        # hide replace stuff
        self.replaceLabel.hide()
        self.replaceEntry.hide()
        self.replaceButton.hide()
        self.replaceAllButton.hide()
        self._replace = False # we are not in replace mode
        visible = self.isVisible()
        if not visible:
            self.showWidget()
        else:
            self.adjustSize()
        cursor = self.currentView().textCursor()
        #if not visible and self.currentView():
        if cursor.hasSelection():
            word = cursor.selection().toPlainText()
            if not re.search(r'\w', word):
                word = ""
            elif self.regexCheck.isChecked():
                word = re.escape(word)
            with qutil.signalsBlocked(self.searchEntry):
                self.searchEntry.setText(word)
            self.slotSearchChanged()
        else:
            self.searchEntry.selectAll()
            self.highlightingOn()
        self.searchEntry.setFocus()

    def replace(self):
        """Called by the main menu Find and Replace... command."""
        # show replace stuff
        self.replaceLabel.show()
        self.replaceEntry.show()
        self.replaceButton.show()
        self.replaceAllButton.show()
        focus = self.replaceEntry if self.isVisible() and self.searchEntry.text() else self.searchEntry
        self._replace = True # we are in replace mode
        if self.isVisible():
            self.adjustSize()
        else:
            self.showWidget()
            self.markPositionsDirty()
            self.updatePositions()
            self.highlightingOn()
        focus.setFocus()

    def slotSearchChanged(self):
        """Called on every change in the search text entry."""
        self._going = True
        self.markPositionsDirty()
        self.updatePositions()
        self.highlightingOn()
        if not self._replace and self._positions:
            positions = [c.position() for c in self._positions]
            cursor = self.currentView().textCursor()
            index = bisect.bisect_left(positions, cursor.selectionStart())
            if index == len(positions):
                index -= 1
            elif index > 0:
                # it might be possible that the text cursor currently already
                # is in a search result. This happens when the search is pop up
                # with an empty text and the current word is then set as search
                # text.
                if cursortools.contains(self._positions[index-1], cursor):
                    index -= 1
            self.gotoPosition(index)
        self._going = False

    def highlightingOn(self, view=None):
        """Show the current search result positions."""
        if view is None:
            view = self.currentView()
        if view:
            viewhighlighter.highlighter(view).highlight("search", self._positions, 1)

    def highlightingOff(self, view=None):
        """Hide the current search result positions."""
        if view is None:
            view = self.currentView()
        if view:
            viewhighlighter.highlighter(view).clear("search")

    def markPositionsDirty(self):
        """Delete positions and mark them dirty, i.e. they need updating."""
        self._positions = []
        self._positionsDirty = True

    def updatePositions(self):
        """Update the search result positions if necessary."""
        view = self.currentView()
        if not view or not self._positionsDirty:
            return
        search = self.searchEntry.text()
        cursor = view.textCursor()
        document = view.document()
        self._positions = []
        if search:
            text = document.toPlainText()
            start = 0
            if (self._replace or not self._going) and cursor.hasSelection():
                # don't search outside the selection
                start = cursor.selectionStart()
                text = text[start:cursor.selectionEnd()]
            flags = re.MULTILINE | re.DOTALL
            if not self.caseCheck.isChecked():
                flags |= re.IGNORECASE
            if not self.regexCheck.isChecked():
                search = re.escape(search)
            try:
                matches = re.finditer(search, text, flags)
            except re.error:
                pass
            else:
                for m in matches:
                    c = QTextCursor(document)
                    c.setPosition(start + m.end())
                    c.setPosition(start + m.start(), QTextCursor.MoveMode.KeepAnchor)
                    self._positions.append(c)
        self.countLabel.setText(format(len(self._positions)))
        enabled = len(self._positions) > 0
        self.replaceButton.setEnabled(enabled)
        self.replaceAllButton.setEnabled(enabled)
        self.prevButton.setEnabled(enabled)
        self.nextButton.setEnabled(enabled)
        self._positionsDirty = False

    def findNext(self):
        """Called on menu Find Next."""
        self._going = True
        self.updatePositions()
        view = self.currentView()
        if view and self._positions:
            positions = [c.position() for c in self._positions]
            index = bisect.bisect_right(positions, view.textCursor().position())
            if index < len(positions):
                self.gotoPosition(index)
            else:
                self.gotoPosition(0)
            view.ensureCursorVisible()
        self._going = False

    def findPrevious(self):
        """Called on menu Find Previous."""
        self._going = True
        self.updatePositions()
        view = self.currentView()
        positions = [c.position() for c in self._positions]
        if view and positions:
            index = bisect.bisect_left(positions, view.textCursor().position()) - 1
            self.gotoPosition(index)
        self._going = False

    def gotoPosition(self, index):
        """Scrolls the current View to the position in the _positions list at index."""
        c = QTextCursor(self._positions[index])
        #c.clearSelection()
        self.currentView().gotoTextCursor(c)
        self.currentView().ensureCursorVisible()

    def event(self, ev):
        """Reimplemented to catch F1 for help and Tab so it does not reach the View."""
        if ev.type() == QEvent.Type.ShortcutOverride and ev.matches(QKeySequence.StandardKey.HelpContents):
            userguide.show("search_replace")
            ev.accept()
            return True
        elif ev.type() == QEvent.Type.KeyPress:
            # This is written this way because Qt.KeyboardModifier,
            # Qt.Modifier, and their corresponding int values are not
            # interchangeable in PyQt6, even though they are in C++.
            modifiers = ev.modifiers().value & (Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Modifier.ALT | Qt.Modifier.META).value
            if ev.key() == Qt.Key.Key_Tab and modifiers == 0:
                # prevent Tab from reaching the View widget
                self.window().focusNextChild()
                ev.accept()
                return True
            elif ev.key() == Qt.Key.Key_Backtab and modifiers & ~Qt.Modifier.SHIFT.value == 0:
                # prevent Tab from reaching the View widget
                self.window().focusPreviousChild()
                ev.accept()
                return True
        return super().event(ev)

    def keyPressEvent(self, ev):
        """Catches Up and Down to jump between search results."""
        # if in search mode, Up and Down jump between search results
        if not self._replace and self._positions and self.searchEntry.text() and not ev.modifiers():
            if ev.key() == Qt.Key.Key_Up:
                self.findPrevious()
                return
            elif ev.key() ==  Qt.Key.Key_Down:
                self.findNext()
                return
        # use enter or return for search next
        if ev.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            self.findNext()
            return
        super().keyPressEvent(ev)

    def doReplace(self, cursor):
        """Perform one replace action."""
        text = cursor.selection().toPlainText()
        search = self.searchEntry.text()
        replace = self.replaceEntry.text()
        ok = text == search
        if self.regexCheck.isChecked():
            m = re.match(search, text)
            ok = False
            if m:
                try:
                    replace = m.expand(replace)
                    ok = True
                except re.error:
                    pass
        if ok:
            pos = cursor.position()
            cursor.insertText(replace)
            cursor.setPosition(pos, QTextCursor.MoveMode.KeepAnchor)
        return ok

    def slotReplace(self):
        """Called when the user clicks Replace."""
        view = self.currentView()
        if view and self._positions:
            positions = [c.position() for c in self._positions]
            index = bisect.bisect_left(positions, view.textCursor().position())
            if index >= len(positions):
                index = 0
            if self.doReplace(self._positions[index]):
                self.findNext()

    def slotReplaceAll(self):
        """Called when the user clicks Replace All."""
        view = self.currentView()
        if view:
            replaced = False
            cursors = self._positions
            if view.textCursor().hasSelection():
                cursors = [cursor for cursor in cursors if cursortools.contains(view.textCursor(), cursor)]
            with cursortools.compress_undo(view.textCursor()):
                for cursor in cursors:
                    if self.doReplace(cursor):
                        replaced = True
            if replaced:
                self.highlightingOn()



# vimode -- Vi Mode package for QPlainTextEdit
#
# Copyright (c) 2012 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Insert ViMode.
"""


from PyQt6.QtCore import Qt

from . import handlerbase


class InsertMode(handlerbase.Handler):

    def enter(self):
        m = self.vimode
        m.textEdit().setCursorWidth(m._originalCursorWidth)
        m.clearCursor()

    def handleKeyPress(self, ev):
        if ev.key() == Qt.Key.Key_Escape and ev.modifiers() == Qt.KeyboardModifier.NoModifier:
            self.setNormalMode()
            return True




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Wood wind part types.
"""


from . import _base
from . import register


class WoodWindPart(_base.SingleVoicePart):
    """Base class for wood wind part types."""


class Flute(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Flute")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Flute", "Fl.")

    midiInstrument = 'flute'


class Piccolo(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Piccolo")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Piccolo", "Pic.")

    midiInstrument = 'piccolo'
    octave = 2
    transposition = (1, 0, 0)


class AltoFlute(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Alto flute")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Alto flute", "A.Fl.")

    midiInstrument = 'flute'
    octave = 0
    transposition = (-1, 4, 0)


class BassFlute(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Bass flute")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Bass flute", "B.Fl.")

    midiInstrument = 'flute'
    octave = 0
    transposition = (-1, 0, 0)


class Oboe(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Oboe")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Oboe", "Ob.")

    midiInstrument = 'oboe'


class OboeDAmore(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Oboe d'amore")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Oboe d'amore", "Ob.D'am.")

    midiInstrument = 'oboe'
    octave = 0
    transposition = (-1, 5, 0)


class EnglishHorn(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("English horn")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for English horn", "Eng.H.")

    midiInstrument = 'english horn'
    octave = 0
    transposition = (-1, 3, 0)


class Bassoon(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Bassoon")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Bassoon", "Bn.")

    midiInstrument = 'bassoon'
    clef = 'bass'
    octave = -1


class ContraBassoon(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Contrabassoon")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Contrabassoon", "C.Bn.")

    midiInstrument = 'bassoon'
    transposition = (-1, 0, 0)
    clef = 'bass'
    octave = -2


class Clarinet(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Clarinet")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Clarinet", "Cl.")

    midiInstrument = 'clarinet'
    transposition = (-1, 6, -1)


class EflatClarinet(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("E-flat clarinet")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for E-flat clarinet", "Cl. in Eb")

    midiInstrument = 'clarinet'
    transposition = (0, 2, -1)


class AClarinet(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("A clarinet")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for A clarinet", "Cl. in A")

    midiInstrument = 'clarinet'
    octave = 0
    transposition = (-1, 5, 0)


class BassClarinet(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Bass clarinet")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Bass clarinet", "B.Cl.")

    midiInstrument = 'clarinet'
    octave = -1
    transposition = (-2, 6, -1)


class C_MelodySax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("C-melody saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for C-melody saxophone", "C-Mel Sax")

    midiInstrument = 'soprano sax'


class SopraninoSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Sopranino saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Sopranino saxophone", "Si.Sax.")

    midiInstrument = 'soprano sax'
    transposition = (0, 2, -1)    # es'


class SopranoSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Soprano saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Soprano saxophone", "So.Sax.")

    midiInstrument = 'soprano sax'
    transposition = (-1, 6, -1)   # bes


class AltoSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Alto saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Alto saxophone", "A.Sax.")

    midiInstrument = 'alto sax'
    octave = 0
    transposition = (-1, 2, -1)   # es


class TenorSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Tenor saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Tenor saxophone", "T.Sax.")

    midiInstrument = 'tenor sax'
    octave = 0
    transposition = (-2, 6, -1)   # bes,


class BaritoneSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Baritone saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Baritone saxophone", "B.Sax.")

    midiInstrument = 'baritone sax'
    octave = -1
    transposition = (-2, 2, -1)   # es,


class BassSax(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Bass saxophone")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Bass saxophone", "Bs.Sax.")

    midiInstrument = 'baritone sax'
    octave = -1
    transposition = (-3, 6, -1)   # bes,,


class SopraninoRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Sopranino recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Sopranino recorder", "Si.Rec.")

    midiInstrument = 'recorder'
    octave = 2
    transposition = (1, 0, 0)

class SopranoRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Soprano recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Soprano recorder", "S.Rec.")

    midiInstrument = 'recorder'
    octave = 2
    transposition = (1, 0, 0)


class AltoRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Alto recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Alto recorder", "A.Rec.")

    midiInstrument = 'recorder'


class TenorRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Tenor recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Tenor recorder", "T.Rec.")

    midiInstrument = 'recorder'


class BassRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Bass recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Bass recorder", "B.Rec.")

    midiInstrument = 'recorder'
    clef = 'bass'
    octave = 0
    transposition = (1, 0, 0)


class ContraBassRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Contrabass recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Contrabass recorder", "Cb.Rec.")

    midiInstrument = 'recorder'
    clef = 'bass'
    octave = -1


class SubContraBassRecorder(WoodWindPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Subcontrabass recorder")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Subcontrabass recorder", "Scb.Rec.")

    midiInstrument = 'recorder'
    clef = 'bass'
    octave = -2
    transposition = (-1, 0, 0)



register(
    lambda: _("Woodwinds"),
    [
        Flute,
        Piccolo,
        AltoFlute,
        BassFlute,
        Oboe,
        OboeDAmore,
        EnglishHorn,
        Bassoon,
        ContraBassoon,
        Clarinet,
        EflatClarinet,
        AClarinet,
        BassClarinet,
        SopraninoSax,
        SopranoSax,
        AltoSax,
        TenorSax,
        BaritoneSax,
        BassSax,
        C_MelodySax,
        SopraninoRecorder,
        SopranoRecorder,
        AltoRecorder,
        TenorRecorder,
        BassRecorder,
        ContraBassRecorder,
        SubContraBassRecorder
    ])



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Plucked string part types.
"""


from PyQt6.QtWidgets import (
    QCheckBox, QComboBox, QCompleter, QGridLayout, QHBoxLayout, QLabel,
    QLineEdit, QSpinBox,
)

import listmodel
import completionmodel
import ly.dom

from . import _base
from . import register


class TablaturePart(_base.Part):
    """Base class for tablature instrument part types."""

    octave = 0
    clef = None
    transposition = None
    tunings = ()    # may contain a list of tunings.
    tabFormat = ''  # can contain a tablatureFormat value.
    midiInstruments = ()  # may contain a list of MIDI instruments.

    def createWidgets(self, layout):
        self.staffTypeLabel = QLabel()
        self.staffType = QComboBox()
        self.staffTypeLabel.setBuddy(self.staffType)
        self.staffType.setModel(listmodel.ListModel(tablatureStaffTypes,
            self.staffType, display=listmodel.translate))
        box = QHBoxLayout()
        layout.addLayout(box)
        box.addWidget(self.staffTypeLabel)
        box.addWidget(self.staffType)
        if self.tunings:
            self.createTuningWidgets(layout)
            self.staffType.activated.connect(self.slotTabEnable)
            self.slotTabEnable(0)
        if self.midiInstruments:
            self.createMidiInstrumentWidgets(layout)

    def createTuningWidgets(self, layout):
        self.tuningLabel = QLabel()
        self.tuning = QComboBox()
        self.tuningLabel.setBuddy(self.tuning)
        tunings = [('', lambda: _("Default"))]
        tunings.extend(self.tunings)
        tunings.append(('', lambda: _("Custom tuning")))
        self.tuning.setModel(listmodel.ListModel(tunings, self.tuning,
            display=listmodel.translate_index(1)))
        self.tuning.setCurrentIndex(1)
        self.customTuning = QLineEdit(enabled=False)
        completionmodel.complete(self.customTuning,
            "scorewiz/completion/plucked_strings/custom_tuning")
        self.tuning.currentIndexChanged.connect(self.slotCustomTuningEnable)
        box = QHBoxLayout()
        layout.addLayout(box)
        box.addWidget(self.tuningLabel)
        box.addWidget(self.tuning)
        layout.addWidget(self.customTuning)

    def createMidiInstrumentWidgets(self, layout):
        self.midiInstrumentLabel = QLabel()
        self.midiInstrumentSelection = QComboBox()
        self.midiInstrumentSelection.addItems(self.midiInstruments)
        self.midiInstrumentSelection.setCurrentIndex(
            self.midiInstruments.index(self.midiInstrument))
        box = QHBoxLayout()
        layout.addLayout(box)
        box.addWidget(self.midiInstrumentLabel)
        box.addWidget(self.midiInstrumentSelection)

    def translateWidgets(self):
        self.staffTypeLabel.setText(_("Staff type:"))
        self.staffType.model().update()
        if self.tunings:
            self.translateTuningWidgets()
        if self.midiInstruments:
            self.translateMidiInstrumentWidgets()

    def translateTuningWidgets(self):
        self.tuningLabel.setText(_("Tuning:"))
        self.customTuning.setToolTip('<qt>' + _(
            "Select custom tuning in the combobox and "
            "enter a custom tuning here, e.g. <code>e, a d g b e'</code>. "
            "Use absolute note names in the same language as you want to use "
            "in your document (by default: \"nederlands\")."))
        try:
            self.customTuning.setPlaceholderText(_("Custom tuning..."))
        except AttributeError:
            pass # only in Qt 4.7+
        self.tuning.model().update()

    def translateMidiInstrumentWidgets(self):
        self.midiInstrumentLabel.setText(_("MIDI instrument:"))

    def slotTabEnable(self, enable):
        """Called when the user changes the staff type.

        Non-zero if the user wants a TabStaff.

        """
        self.tuning.setEnabled(bool(enable))
        if enable:
            self.slotCustomTuningEnable(self.tuning.currentIndex())
        else:
            self.customTuning.setEnabled(False)

    def slotCustomTuningEnable(self, index):
        self.customTuning.setEnabled(index > len(self.tunings))

    def voiceCount(self):
        """Returns the number of voices.

        Inherit to make this user-settable.

        """
        return 1

    def build(self, data, builder):
        # First make assignments for the voices we want to create
        numVoices = self.voiceCount()
        if numVoices == 1:
            voices = (ly.util.mkid(data.name()),)
        elif numVoices == 2:
            order = 1, 2
            voices = 'upper', 'lower'
        elif numVoices == 3:
            order = 1, 3, 2
            voices = 'upper', 'middle', 'lower'
        else:
            order = 1, 2, 3, 4
            voices = [ly.util.mkid(data.name(), "voice") + ly.util.int2text(i) for i in order]

        assignments = [data.assignMusic(name, self.octave) for name in voices]

        staffType = self.staffType.currentIndex()
        if staffType in (0, 2):
            # create a normal staff
            staff = ly.dom.Staff()
            seq = ly.dom.Seqr(staff)
            if self.clef:
                ly.dom.Clef(self.clef, seq)
            if self.transposition is not None:
                seq = builder.setStaffTransposition(seq, self.transposition)
            mus = ly.dom.Simr(seq)
            for a in assignments[:-1]:
                ly.dom.Identifier(a.name, mus)
                ly.dom.VoiceSeparator(mus)
            ly.dom.Identifier(assignments[-1].name, mus)
            if self.midiInstruments:
                builder.setMidiInstrument(staff,
                    self.midiInstrumentSelection.currentText())
            else:
                builder.setMidiInstrument(staff, self.midiInstrument)

        if staffType in (1, 2):
            # create a tab staff
            tabstaff = ly.dom.TabStaff()
            if self.tabFormat:
                tabstaff.getWith()['tablatureFormat'] = ly.dom.Scheme(self.tabFormat)
            self.setTunings(tabstaff)
            if self.midiInstruments:
                builder.setMidiInstrument(tabstaff,
                    self.midiInstrumentSelection.currentText())
            else:
                builder.setMidiInstrument(tabstaff, self.midiInstrument)
            sim = ly.dom.Simr(tabstaff)
            if numVoices == 1:
                ly.dom.Identifier(assignments[0].name, sim)
            else:
                for num, a in zip(order, assignments):
                    s = ly.dom.Seq(ly.dom.TabVoice(parent=sim))
                    ly.dom.Text('\\voice' + ly.util.int2text(num), s)
                    ly.dom.Identifier(a.name, s)

        if staffType == 0:
            # only a normal staff
            p = staff
        elif staffType == 1:
            # only a TabStaff
            p = tabstaff
        else:
            # both TabStaff and normal staff
            p = ly.dom.StaffGroup()
            s = ly.dom.Sim(p)
            s.append(staff)
            s.append(tabstaff)
            p.getWith()['systemStartDelimiter'] = ly.dom.Scheme("'SystemStartSquare")

        builder.setInstrumentNamesFromPart(p, self, data)
        data.nodes.append(p)

    def setTunings(self, tab):
        if self.tunings:
            i = self.tuning.currentIndex()
            if i == 0:
                return
            elif i > len(self.tunings):
                value = ly.dom.Text(f"\\stringTuning <{self.customTuning.text()}>")
            else:
                tuning = self.tunings[self.tuning.currentIndex() - 1][0]
                value = ly.dom.Scheme(tuning)
            tab.getWith()['stringTunings'] = value


tablatureStaffTypes = (
    lambda: _("Normal staff"),
    lambda: _("Tablature"),
    #L10N: Both a Normal and a Tablature staff
    lambda: _("Both"),
)


class Mandolin(TablaturePart):
    @staticmethod
    def title(_=_base.translate):
        return _("Mandolin")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Mandolin", "Mdl.")

    midiInstrument = 'acoustic guitar (steel)'
    tunings = (
        ('mandolin-tuning', lambda: _("Mandolin tuning")),
    )


class Ukulele(TablaturePart):
    @staticmethod
    def title(_=_base.translate):
        return _("Ukulele")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Ukulele", "Uk.")

    midiInstrument = 'acoustic guitar (nylon)'
    tunings = (
        ('ukulele-tuning', lambda: _("Ukulele tuning")),
        ('ukulele-d-tuning', lambda: _("Ukulele D-tuning")),
        ('tenor-ukulele-tuning', lambda: _("Tenor Ukulele tuning")),
        ('baritone-ukulele-tuning', lambda: _("Baritone Ukulele tuning")),
    )


class Banjo(TablaturePart):
    @staticmethod
    def title(_=_base.translate):
        return _("Banjo")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Banjo", "Bj.")

    midiInstrument = 'banjo'
    transposition = (-1, 0, 0)
    tabFormat = 'fret-number-tablature-format-banjo'
    tunings = (
        ('banjo-open-g-tuning', lambda: _("Open G-tuning (aDGBD)")),
        ('banjo-c-tuning', lambda: _("C-tuning (gCGBD)")),
        ('banjo-modal-tuning', lambda: _("Modal tuning (gDGCD)")),
        ('banjo-open-d-tuning', lambda: _("Open D-tuning (aDF#AD)")),
        ('banjo-open-dm-tuning', lambda: _("Open Dm-tuning (aDFAD)")),
    )

    def createTuningWidgets(self, layout):
        super().createTuningWidgets(layout)
        self.fourStrings = QCheckBox()
        layout.addWidget(self.fourStrings)

    def translateTuningWidgets(self):
        super().translateTuningWidgets()
        self.fourStrings.setText(_("Four strings (instead of five)"))

    def setTunings(self, tab):
        i = self.tuning.currentIndex()
        if i > len(self.tunings) or not self.fourStrings.isChecked():
            super().setTunings(tab)
        else:
            tab.getWith()['stringTunings'] = ly.dom.Scheme(
                '(four-string-banjo {})'.format(
                    self.tunings[i][0]))

    def slotCustomTuningEnable(self, index):
        super().slotCustomTuningEnable(index)
        self.fourStrings.setEnabled(index <= len(self.tunings))


class Guitar(TablaturePart):
    @staticmethod
    def title(_=_base.translate):
        return _("Guitar")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Guitar", "Gt.")

    midiInstrument = 'acoustic guitar (nylon)'
    transposition = (-1, 0, 0)  # but see build() below
    tunings = (
        ('guitar-tuning', lambda: _("Guitar tuning")),
        ('guitar-seven-string-tuning', lambda: _("Guitar seven-string tuning")),
        ('guitar-drop-d-tuning', lambda: _("Guitar drop-D tuning")),
        ('guitar-drop-c-tuning', lambda: _("Guitar drop-C tuning")),
        ('guitar-open-g-tuning', lambda: _("Open G-tuning")),
        ('guitar-open-d-tuning', lambda: _("Guitar open D tuning")),
        ('guitar-dadgad-tuning', lambda: _("Guitar d-a-d-g-a-d tuning")),
        ('guitar-lute-tuning', lambda: _("Lute tuning")),
        ('guitar-asus4-tuning', lambda: _("Guitar A-sus4 tuning")),
    )
    def createWidgets(self, layout):
        super().createWidgets(layout)
        self.voicesLabel = QLabel()
        self.voices = QSpinBox(minimum=1, maximum=4, value=1)
        box = QHBoxLayout()
        box.addWidget(self.voicesLabel)
        box.addWidget(self.voices)
        layout.addLayout(box)
        self.octaveClef = QCheckBox()
        layout.addWidget(self.octaveClef)
        # Use treble_8 by default for (classical) Guitar but not derived
        # types, since it is less common in the styles those are intended for
        self.octaveClef.setChecked(type(self) is Guitar)

    def translateWidgets(self):
        super().translateWidgets()
        self.voicesLabel.setText(_("Voices:"))
        self.octaveClef.setText(_("Include octave indication"))
        self.octaveClef.setToolTip(_(
            "Use an octave (treble_8) clef to make transposition explicit "
            "as preferred by some style guides."))

    def voiceCount(self):
        return self.voices.value()

    def build(self, data, builder):
        if self.octaveClef.isChecked():
            self.clef = "treble_8"
            self.transposition = None
        else:
            self.clef = None
            self.transposition = (-1, 0, 0)
        super().build(data, builder)


class AcousticGuitar(Guitar):
    @staticmethod
    def title(_=_base.translate):
        return _("Acoustic guitar")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Acoustic guitar", "A.Gt.")

    midiInstrument = 'acoustic guitar (steel)'
    midiInstruments = (
        'acoustic guitar (nylon)',
        'acoustic guitar (steel)',
    )


class ElectricGuitar(Guitar):
    @staticmethod
    def title(_=_base.translate):
        return _("Electric guitar")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Electric guitar", "E.Gt.")

    midiInstrument = 'electric guitar (clean)'
    midiInstruments = (
        'electric guitar (jazz)',
        'electric guitar (clean)',
        'electric guitar (muted)',
        'overdriven guitar',
        'distorted guitar',
    )


class AcousticBass(TablaturePart):
    @staticmethod
    def title(_=_base.translate):
        return _("Acoustic bass")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Acoustic bass", "A.Bs.") #FIXME

    midiInstrument = 'acoustic bass'
    clef = 'bass'
    octave = -2
    transposition = (-1, 0, 0)
    tunings = (
        ('bass-tuning', lambda: _("Bass tuning")),
        ('bass-four-string-tuning', lambda: _("Four-string bass tuning")),
        ('bass-drop-d-tuning', lambda: _("Bass drop-D tuning")),
        ('bass-five-string-tuning', lambda: _("Five-string bass tuning")),
        ('bass-six-string-tuning', lambda: _("Six-string bass tuning")),
    )


class ElectricBass(AcousticBass):
    @staticmethod
    def title(_=_base.translate):
        return _("Electric bass")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Electric bass", "E.Bs.")

    midiInstrument = 'electric bass (finger)'
    midiInstruments = (
        'electric bass (finger)',
        'electric bass (pick)',
        'fretless bass',
        'slap bass 1',
        'slap bass 2',
    )


class Harp(_base.PianoStaffPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Harp")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Harp", "Hp.")

    midiInstrument = 'orchestral harp'

    def translateWidgets(self):
        super().translateWidgets()
        self.upperVoicesLabel.setText(_("Upper staff:"))
        self.lowerVoicesLabel.setText(_("Lower staff:"))

    def build(self, data, builder):
        p = ly.dom.PianoStaff()
        builder.setInstrumentNamesFromPart(p, self, data)
        s = ly.dom.Sim(p)
        # add two staves, with a respective number of voices.
        self.buildStaff(data, builder, 'upper', 1, self.upperVoices.value(), s)
        self.buildStaff(data, builder, 'lower', 0, self.lowerVoices.value(), s, "bass")
        data.nodes.append(p)





register(
    lambda: _("Plucked strings"),
    [
        Mandolin,
        Banjo,
        Ukulele,
        Guitar,
        AcousticGuitar,
        ElectricGuitar,
        AcousticBass,
        ElectricBass,
        Harp,
    ])




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Import Music XML dialog.
Uses musicxml2ly to create ly file from xml.
In the dialog the options of musicxml2ly can be set.
"""


from PyQt6.QtCore import QSettings, QSize
from PyQt6.QtWidgets import (QCheckBox, QComboBox, QDialogButtonBox, QLabel)

import app
import qutil
import job

from . import toly_dialog

# language names musicxml2ly allows
_langlist = [
    'nederlands',
    'catalan',
    'deutsch',
    'english',
    'espanol',
    'italiano',
    'norsk',
    'portugues',
    'suomi',
    'svenska',
    'vlaams',
]


class Dialog(toly_dialog.ToLyDialog):

    def __init__(self, parent=None):

        self.noartCheck = QCheckBox()
        self.norestCheck = QCheckBox()
        self.nolayoutCheck = QCheckBox()
        self.nobeamCheck = QCheckBox()
        self.useAbsCheck = QCheckBox()
        self.commMidiCheck = QCheckBox()

        self.langCombo = QComboBox()
        self.langLabel = QLabel()

        self.impChecks = [self.noartCheck,
                          self.norestCheck,
                          self.nolayoutCheck,
                          self.nobeamCheck,
                          self.useAbsCheck,
                          self.commMidiCheck]

        self.noartCheck.setObjectName("articulation-directions")
        self.norestCheck.setObjectName("rest-positions")
        self.nolayoutCheck.setObjectName("page-layout")
        self.nobeamCheck.setObjectName("import-beaming")
        self.useAbsCheck.setObjectName("absolute-mode")
        self.commMidiCheck.setObjectName("comment-out-midi")

        self.langCombo.addItem('')
        self.langCombo.addItems(_langlist)

        self.impExtra = [self.langLabel, self.langCombo]

        super().__init__(
            parent,
            imp_prgm="musicxml2ly",
            userg="musicxml_import")

        app.translateUI(self)
        qutil.saveDialogSize(self, "musicxml_import/dialog/size", QSize(480, 800))

        self.loadSettings()

    def translateUI(self):
        self.setWindowTitle(app.caption(_("Import Music XML")))
        self.noartCheck.setText(_("Import articulation directions"))
        self.norestCheck.setText(_("Import rest positions"))
        self.nolayoutCheck.setText(_("Import page layout"))
        self.nobeamCheck.setText(_("Import beaming"))
        self.useAbsCheck.setText(_("Pitches in absolute mode"))
        self.commMidiCheck.setText(_("Comment out midi block"))

        self.langLabel.setText(_("Language for pitch names"))
        self.langCombo.setItemText(0, _("Default"))

        self.buttons.button(QDialogButtonBox.StandardButton.Ok).setText(_("Run musicxml2ly"))

        super().translateUI()

    def configure_job(self):
        super().configure_job()
        j = self._job
        if self.useAbsCheck.isChecked():
            j.add_argument('-a')
        if not self.noartCheck.isChecked():
            j.add_argument('--nd')
        if not self.norestCheck.isChecked():
            j.add_argument('--nrp')
        if not self.nolayoutCheck.isChecked():
            j.add_argument('--npl')
        if not self.nobeamCheck.isChecked():
            j.add_argument('--no-beaming')
        if not self.commMidiCheck.isChecked():
            j.add_argument('-m')
        index = self.langCombo.currentIndex()
        if index > 0:
            j.add_argument('--language=' + _langlist[index - 1])

    def loadSettings(self):
        """Get users previous settings."""
        self.imp_default = [False, False, False, False, False, False]
        self.settings = QSettings()
        self.settings.beginGroup('musicxml_import')
        super().loadSettings()
        lang = self.settings.value("language", "default", str)
        try:
            index = _langlist.index(lang)
        except ValueError:
            index = -1
        self.langCombo.setCurrentIndex(index + 1)

    def saveSettings(self):
        """Save users last settings."""
        self.settings = QSettings()
        self.settings.beginGroup('musicxml_import')
        super().saveSettings()
        index = self.langCombo.currentIndex()
        self.settings.setValue('language', 'default' if index == 0 else _langlist[index-1])



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Base ExtensionWidget for extensions
"""

from PyQt6.QtCore import Qt, QObject
from PyQt6.QtWidgets import QLabel, QVBoxLayout, QWidget

import app

from . import ExtensionMixin

class FailedExtensionWidget(QWidget):
    """Simple placeholder widget that is used if an extension's
    Tool Panel fails to instantiate."""

    def __init__(self, parent):
        super().__init__(parent)
        self.label = QLabel()
        self.label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        layout = QVBoxLayout()
        layout.addStretch()
        layout.addWidget(self.label)
        layout.addStretch()
        self.setLayout(layout)
        self.translateUI()

    def translateUI(self):
        self.label.setText(_("Tool Panel failed to load"))


class ExtensionWidget(QWidget, ExtensionMixin):
    """Convenience base class for widgets in an extension.
    Can be used as the immediate super class for a widget to be used in
    an extension that otherwise would be a QWidget. If the widget is
    something else use ExtensionMixin as secondary super class
    instead.

    The class provides additional attributes to directly interact with the
    elements of an extension, see the docstring for ExtensionMixin
    for details.
    """

    pass



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Music View preferences.
"""


import platform
import re

from PyQt6.QtCore import QSettings, Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (
    QAbstractItemView, QButtonGroup, QCheckBox, QComboBox, QDoubleSpinBox,
    QFontComboBox, QGridLayout, QHBoxLayout, QLabel, QPushButton, QRadioButton,
    QSlider, QSpinBox, QVBoxLayout, QWidget)

import app
import qutil
import preferences
import pagedview
import qpageview.constants


class MusicViewers(preferences.ScrolledGroupsPage):
    def __init__(self, dialog):
        super().__init__(dialog)

        layout = QVBoxLayout()
        self.scrolledWidget.setLayout(layout)

        layout.addWidget(Documents(self))
        layout.addWidget(PageScaling(self))
        layout.addWidget(PageLayout(self))
        layout.addWidget(Scrolling(self))
        layout.addWidget(ViewerOptions(self))
        layout.addWidget(Magnifier(self))
        layout.addWidget(Printing(self))
        layout.addStretch(1)


class Documents(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.newerFilesOnly = QCheckBox(toggled=self.changed)
        layout.addWidget(self.newerFilesOnly)

        self.documentProperties = QCheckBox(toggled=self.changed)
        layout.addWidget(self.documentProperties)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Documents"))
        self.newerFilesOnly.setText(_("Only load updated PDF documents"))
        self.newerFilesOnly.setToolTip(_(
            "If checked, Frescobaldi will not open PDF documents that are not\n"
            "up-to-date (i.e. the source file has been modified later)."))
        self.documentProperties.setText(_("Remember View settings per-document"))
        self.documentProperties.setToolTip(_(
            "If checked, every document in the Music View will remember its\n"
            "own layout setting, zoom factor, etc. If unchecked, the View will\n"
            "not change its settings when a different document is displayed."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        self.newerFilesOnly.setChecked(s.value("newer_files_only", True, bool))
        self.documentProperties.setChecked(s.value("document_properties", True, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        s.setValue("newer_files_only", self.newerFilesOnly.isChecked())
        s.setValue("document_properties", self.documentProperties.isChecked())


class PageScaling(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QGridLayout()
        self.setLayout(layout)

        self.viewModeGroup = QButtonGroup()

        self.viewFixedScale = QRadioButton(toggled=self.changed)
        self.viewModeGroup.addButton(self.viewFixedScale)
        self.viewModeGroup.setId(self.viewFixedScale, qpageview.constants.FixedScale)
        layout.addWidget(self.viewFixedScale, 0, 0, 1, 1)

        self.viewFitHeight = QRadioButton(toggled=self.changed)
        self.viewModeGroup.addButton(self.viewFitHeight)
        self.viewModeGroup.setId(self.viewFitHeight, qpageview.constants.FitHeight)
        layout.addWidget(self.viewFitHeight, 1, 0, 1, 1)

        self.viewFitWidth = QRadioButton(toggled=self.changed)
        self.viewModeGroup.addButton(self.viewFitWidth)
        self.viewModeGroup.setId(self.viewFitWidth, qpageview.constants.FitWidth)
        layout.addWidget(self.viewFitWidth, 2, 0, 1, 1)

        self.viewFitBoth = QRadioButton(toggled=self.changed)
        self.viewModeGroup.addButton(self.viewFitBoth)
        self.viewModeGroup.setId(self.viewFitBoth, qpageview.constants.FitBoth)
        layout.addWidget(self.viewFitBoth, 3, 0, 1, 1)

        maxScale = int(pagedview.PagedView.MAX_ZOOM * 100)
        self.initialScaleSlider = QSlider(Qt.Orientation.Horizontal, valueChanged=self.changed)
        self.initialScaleSlider.setSingleStep(50)
        self.initialScaleSlider.setRange(50, maxScale)
        self.initialScaleSpinBox = QSpinBox()
        self.initialScaleSpinBox.setRange(50, maxScale)
        self.initialScaleSpinBox.valueChanged.connect(self.initialScaleSlider.setValue)
        self.initialScaleSlider.valueChanged.connect(self.initialScaleSpinBox.setValue)
        layout.addWidget(self.initialScaleSlider, 0, 1, 1, 2)
        layout.addWidget(self.initialScaleSpinBox, 0, 3)

        self.viewFixedScale.toggled.connect(self.toggleFixedScaleControls)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Page scaling"))
        self.viewFixedScale.setText(_("Fixed scale:"))
        self.viewFitHeight.setText(_("Fit height"))
        self.viewFitWidth.setText(_("Fit width"))
        self.viewFitBoth.setText(_("Fit page")) # to match the Music menu
        self.initialScaleSpinBox.setSuffix(_("percent unit sign", "%"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        # These are from qpageview.view.ViewProperties
        v = s.value("viewMode", qpageview.constants.FixedScale, int)
        b = self.viewModeGroup.button(v)
        if b:
            b.setChecked(True)
        self.toggleFixedScaleControls(v == qpageview.constants.FixedScale)
        self.initialScaleSlider.setValue(round(s.value("zoomFactor", 1.0, float) * 100))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        # These are from qpageview.view.ViewProperties
        s.setValue("viewMode", self.viewModeGroup.checkedId())
        s.setValue("zoomFactor", self.initialScaleSlider.value() / 100.0)

    def toggleFixedScaleControls(self, enabled):
        self.initialScaleSlider.setEnabled(enabled)
        self.initialScaleSpinBox.setEnabled(enabled)


class PageLayout(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.pageLayoutGroup = QButtonGroup()

        self.pageLayoutSingle = QRadioButton(toggled=self.changed)
        self.pageLayoutGroup.addButton(self.pageLayoutSingle)
        layout.addWidget(self.pageLayoutSingle)

        self.pageLayoutDoubleRight = QRadioButton(toggled=self.changed)
        self.pageLayoutGroup.addButton(self.pageLayoutDoubleRight)
        layout.addWidget(self.pageLayoutDoubleRight)

        self.pageLayoutDoubleLeft = QRadioButton(toggled=self.changed)
        self.pageLayoutGroup.addButton(self.pageLayoutDoubleLeft)
        layout.addWidget(self.pageLayoutDoubleLeft)

        self.pageLayoutRaster = QRadioButton(toggled=self.changed)
        self.pageLayoutGroup.addButton(self.pageLayoutRaster)
        layout.addWidget(self.pageLayoutRaster)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Page layout"))
        self.pageLayoutSingle.setText(_("Single"))
        self.pageLayoutDoubleRight.setText(_("Two pages (first page right)"))
        self.pageLayoutDoubleLeft.setText(_("Two pages (first page left)"))
        self.pageLayoutRaster.setText(_("Grid layout"))
        self.pageLayoutRaster.setToolTip(_("The layout of pages (horizontal or vertical) adjusts dynamically based on the zoom level and the available space in the Music View. Continuous scrolling option must be checked."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        # These are from qpageview.view.ViewProperties
        v = s.value("pageLayoutMode", "single")
        if v == "single":
            self.pageLayoutSingle.setChecked(True)
        elif v == "double_right":
            self.pageLayoutDoubleRight.setChecked(True)
        elif v == "double_left":
            self.pageLayoutDoubleLeft.setChecked(True)
        elif v == "raster":
            self.pageLayoutRaster.setChecked(True)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        # These are from qpageview.view.ViewProperties
        if self.pageLayoutSingle.isChecked():
            s.setValue("pageLayoutMode", "single")
        elif self.pageLayoutDoubleRight.isChecked():
            s.setValue("pageLayoutMode", "double_right")
        elif self.pageLayoutDoubleLeft.isChecked():
            s.setValue("pageLayoutMode", "double_left")
        elif self.pageLayoutRaster.isChecked():
            s.setValue("pageLayoutMode", "raster")


class Scrolling(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QGridLayout()
        self.setLayout(layout)

        self.orientationLabel = QLabel()
        layout.addWidget(self.orientationLabel, 0, 0)
        self.orientationGroup = QButtonGroup()
        self.orientationHorizontal = QRadioButton(toggled=self.changed)
        self.orientationGroup.addButton(self.orientationHorizontal)
        self.orientationGroup.setId(self.orientationHorizontal, qpageview.constants.Horizontal)
        layout.addWidget(self.orientationHorizontal, 0, 1)
        self.orientationVertical = QRadioButton(toggled=self.changed)
        self.orientationGroup.addButton(self.orientationVertical)
        self.orientationGroup.setId(self.orientationVertical, qpageview.constants.Vertical)
        layout.addWidget(self.orientationVertical, 0, 2)

        self.continuousMode = QCheckBox(toggled=self.changed)
        layout.addWidget(self.continuousMode, 1, 0, 1, 4)

        self.enableKineticScrolling = QCheckBox(toggled=self.changed)
        layout.addWidget(self.enableKineticScrolling, 2, 0, 1, 4)

        self.enableStrictPaging = QCheckBox(toggled=self.changed)
        layout.addWidget(self.enableStrictPaging, 3, 0, 1, 4)

        app.translateUI(self)

    def translateUI(self):
        # L10N: "Kinetic Scrolling" is a checkbox label, as in "Enable Kinetic Scrolling"
        self.setTitle(_("Scrolling"))
        self.orientationLabel.setText(_("Orientation:"))
        self.orientationHorizontal.setText(_("Horizontal"))
        self.orientationVertical.setText(_("Vertical"))
        self.continuousMode.setText(_("Continuous scrolling"))
        self.enableKineticScrolling.setText(_("Kinetic scrolling"))
        self.enableStrictPaging.setText(
            _("Use Page Up and Page Down keys to change pages"))
        self.enableStrictPaging.setToolTip(_(
            "If checked, PageUp and PageDown always page to the previous or next page instead of scrolling."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        kineticScrollingActive = s.value("kinetic_scrolling", True, bool)
        self.enableKineticScrolling.setChecked(kineticScrollingActive)
        strictPaging = s.value("strict_paging", False, bool)
        self.enableStrictPaging.setChecked(strictPaging)
        # These are from qpageview.view.ViewProperties
        v = s.value("orientation", qpageview.constants.Vertical, int)
        b = self.orientationGroup.button(v)
        if b:
            b.setChecked(True)
        self.continuousMode.setChecked(s.value("continuousMode", True, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        s.setValue("kinetic_scrolling", self.enableKineticScrolling.isChecked())
        s.setValue("strict_paging", self.enableStrictPaging.isChecked())
        # These are from qpageview.view.ViewProperties
        s.setValue("orientation", self.orientationGroup.checkedId())
        s.setValue("continuousMode", self.continuousMode.isChecked())


class ViewerOptions(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.showScrollbars = QCheckBox(toggled=self.changed)
        layout.addWidget(self.showScrollbars)

        self.showShadow = QCheckBox(toggled=self.changed)
        layout.addWidget(self.showShadow)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Viewer options"))
        self.showScrollbars.setText(_("Show scrollbars"))
        self.showShadow.setText(_("Show shadow under pages"))
        self.showShadow.setToolTip(_(
            "If checked, Frescobaldi draws a shadow around the pages."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        showScrollbars = s.value("show_scrollbars", True, bool)
        self.showScrollbars.setChecked(showScrollbars)
        shadow = s.value("shadow", True, bool)
        self.showShadow.setChecked(shadow)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        s.setValue("show_scrollbars", self.showScrollbars.isChecked())
        s.setValue("shadow", self.showShadow.isChecked())


class Magnifier(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QGridLayout()
        self.setLayout(layout)

        self.magnifierSizeLabel = QLabel()
        self.magnifierSizeSlider = QSlider(Qt.Orientation.Horizontal, valueChanged=self.changed)
        self.magnifierSizeSlider.setSingleStep(50)
        self.magnifierSizeSlider.setRange(pagedview.Magnifier.MIN_SIZE, pagedview.Magnifier.MAX_SIZE)
        self.magnifierSizeSpinBox = QSpinBox()
        self.magnifierSizeSpinBox.setRange(pagedview.Magnifier.MIN_SIZE, pagedview.Magnifier.MAX_SIZE)
        self.magnifierSizeSpinBox.valueChanged.connect(self.magnifierSizeSlider.setValue)
        self.magnifierSizeSlider.valueChanged.connect(self.magnifierSizeSpinBox.setValue)
        layout.addWidget(self.magnifierSizeLabel, 0, 0)
        layout.addWidget(self.magnifierSizeSlider, 0, 1, 1, 2)
        layout.addWidget(self.magnifierSizeSpinBox, 0, 3)

        self.magnifierScaleLabel = QLabel()
        self.magnifierScaleSlider = QSlider(Qt.Orientation.Horizontal, valueChanged=self.changed)
        self.magnifierScaleSlider.setSingleStep(50)
        self.magnifierScaleSlider.setRange(50, 800)
        self.magnifierScaleSpinBox = QSpinBox()
        self.magnifierScaleSpinBox.setRange(50, 800)
        self.magnifierScaleSpinBox.valueChanged.connect(self.magnifierScaleSlider.setValue)
        self.magnifierScaleSlider.valueChanged.connect(self.magnifierScaleSpinBox.setValue)
        layout.addWidget(self.magnifierScaleLabel, 1, 0)
        layout.addWidget(self.magnifierScaleSlider, 1, 1, 1, 2)
        layout.addWidget(self.magnifierScaleSpinBox, 1, 3)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Magnifier"))
        self.magnifierSizeLabel.setText(_("Size:"))
        self.magnifierSizeLabel.setToolTip(_(
            "Size of the magnifier glass (Ctrl+Click in the Music View)."))
        # L10N: as in "400 pixels", appended after number in spinbox, note the leading space
        self.magnifierSizeSpinBox.setSuffix(_(" pixels"))
        self.magnifierScaleLabel.setText(_("Scale:"))
        self.magnifierScaleLabel.setToolTip(_(
            "Magnification of the magnifier."))
        self.magnifierScaleSpinBox.setSuffix(_("percent unit sign", "%"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        self.magnifierSizeSlider.setValue(s.value("magnifier/size", 350, int))
        self.magnifierScaleSlider.setValue(round(s.value("magnifier/scalef", 3.0, float) * 100))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("musicview")
        s.setValue("magnifier/size", self.magnifierSizeSlider.value())
        s.setValue("magnifier/scalef", self.magnifierScaleSlider.value() / 100.0)


class Printing(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QGridLayout()
        self.setLayout(layout)
        self.resolutionLabel = QLabel()
        self.resolution = QComboBox(editable=True, editTextChanged=page.changed)
        self.resolution.addItems("300 600 1200".split())
        self.resolution.lineEdit().setInputMask("9000")

        layout.addWidget(self.resolutionLabel, 0, 0)
        layout.addWidget(self.resolution, 0, 1)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Printing"))
        self.resolutionLabel.setText(_("Resolution:"))
        self.resolution.setToolTip(_(
            "Set the resolution to be sent to the printer."))

    def loadSettings(self):
        s = QSettings()
        with qutil.signalsBlocked(self.resolution):
            self.resolution.setEditText(format(s.value("printing/dpi", 300, int)))

    def saveSettings(self):
        s = QSettings()
        s.setValue("printing/dpi", int(self.resolution.currentText()))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Base ExtensionActionCollection for extensions
"""

from PyQt6.QtWidgets import QMenu

import actioncollection

class ExtensionActionCollection(actioncollection.ActionCollection):
    """An ActionCollection descendant for use with extensions.
    Apart from the added functionality it's noteworthy that
    the 'name' variable is emptied and that the collection is local
    to the extension (i.e. not used in the globa action collection
    manager).
    """

    name = ""

    def __init__(self, extension):
        self._extension = extension
        super().__init__(extension.mainwindow())
        # Initialize menu actions.
        # By default the Tools menu entry is set to None
        # (causing all actions to be used) while those for
        # the editor and music viewer are empty lists.
        # By overriding configure_menu_actions() any of the menus
        # can be configured manually.
        self._action_lists = {
            'tools': None,
            'editor': [],
            'musicview': [],
            'manuscriptview': []
        }
        # Interface for subclasses to override menus
        self.configure_menu_actions()
        # Interface for subclasses to handle actions
        self.connect_actions()
        # Interface for initializing actions
        self.load_settings()

    def by_text(self):
        """Returns a list with all actions, sorted by the display text,
        ignoring the & mmemonic character."""
        result = list(self._actions.values())
        result.sort(
            key=lambda action: ''.join(c for c in action.text() if c != "&"))
        return result

    def configure_menu_actions(self):
        """Can be used in subclasses to configure the menu action behaviour.
        Calling self.set_menu_action_list(target, actions) each of the menus
        can be configured separately (see comments there)."""
        pass

    def connect_actions(self):
        """Should be implemented when actions have to be connected
        globally during the extension's lifetime. Alternatively
        the actions can also
        """
        pass

    def extension(self):
        return self._extension

    def load_settings(self):
        """Should be implemented to initialize the actions."""
        pass

    def menu_actions(self, target):
        """Returns a list of actions for use in the extension's entry
        in a menu. By default this is a list of all available
        actions, sorted by the display text. For custom sorting or
        a custom selection this method can be overridden."""
        action_list = self._action_lists.get(target, None)
        if action_list is None:
            action_list = self._action_lists[target] = self.by_text()
        return action_list

    def set_menu_action_list(self, target, actions):
        """Specify the contents of a menu.
        Each entry in the _action_lists dictionary can be None or a list.
        If it is None then the self.by_text() method is used to create a flat
        list of all actions, sorted by display name (default for Tools menu). Otherwise a list of actions or submenus can be stored. The context menus
        are initialized to empty lists, meaning by default an extension does
        *not* get an entry in any context menu.
        """
        self._action_lists[target] = actions

    def settings(self):
        """Reference to the extension's settings() object."""
        return self.extension().settings()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Handle everything around (available) text and music fonts.
NOTE: "available" refers to the fonts that are available to LilyPond,
which may be different than for arbitrary programs and can canonically
be determined by running `lilypond -dshow-available-fonts`.
"""


from PyQt6.QtCore import (
    QObject,
    QSettings,
    QSize
)
from PyQt6.QtGui import QAction

import actioncollection
import actioncollectionmanager
import app
import icons
import plugin
import qutil

from . import musicfonts, textfonts


def fonts(mainwindow):
    return Fonts.instance(mainwindow)


class Fonts(plugin.MainWindowPlugin):
    """
    The main plugin for handling font related actions.
    """

    def __init__(self, mainwindow):
        ac = self.actionCollection = Actions()
        actioncollectionmanager.manager(mainwindow).addActionCollection(ac)
        ac.fonts_document_fonts.triggered.connect(self.document_fonts)

    def document_fonts(self):
        """
        Menu action Document Fonts.
        Depending on the LilyPond version associated with the current document
        show either the new or the old document font dialog.
        """
        mainwin = self.mainwindow()
        view = mainwin.currentView()
        doc = mainwin.currentDocument()
        import documentinfo
        info = documentinfo.lilyinfo(doc)

        if info.version() >= (2, 19, 12):
            # Show the new Document Fonts dialog
            from . import dialog
            dlg = dialog.FontsDialog(mainwin, lilypond_info=info)
            qutil.saveDialogSize(
                dlg, "document-fonts-dialog/dialog-size",
                QSize(1024, 700)
            )
            dlg.exec()
            if dlg.result:
                # is populated when the user has clicked "Use"
                cmd = (
                    dlg.result
                    if dlg.result[-1] == '\n'
                    else dlg.result + '\n'
                )
                view.textCursor().insertText(cmd)
        else:
            # Show the old "Set Document Fonts" dialog
            from . import oldfontsdialog
            dlg = oldfontsdialog.DocumentFontsDialog(self.mainwindow())
            if dlg.exec():
                cmd = dlg.document_font_code()
                # NOTE: How to translate this to the dialog context?
                # if state[-1] != "paper":
                cmd = f"\\paper {{\n{cmd}}}\n"
                view.textCursor().insertText(cmd)


class Actions(actioncollection.ActionCollection):
    name = "fonts"

    def createActions(self, parent=None):
        self.fonts_document_fonts = QAction(parent)
        self.fonts_document_fonts.setIcon(
            icons.get('preferences-desktop-font'))

    def translateUI(self):
        self.fonts_document_fonts.setText(
            _("&Document Fonts..."))
        self.fonts_document_fonts.setToolTip(
            _("Show and select text and music fonts available in the " +
              "LilyPond version of the current document"))


class AvailableFonts(QObject):
    """Store available text and music fonts for a given
    LilyPond installation. Sets for multiple registered
    installations are maintained in a global dictionary
    cached for the whole application lifetime.
    Note that the music fonts are loaded immediately while
    text fonts have to be loaded explicitly (in an
    asynchronous invocation of LilyPond)."""

    def __init__(self, lilypond_info):
        super().__init__()
        self.lilypond_info = lilypond_info
        self._music_fonts = musicfonts.InstalledMusicFonts(
            lilypond_info)
        self._text_fonts = textfonts.TextFonts(lilypond_info)

    def music_fonts(self):
        return self._music_fonts

    def text_fonts(self):
        return self._text_fonts


# Caches AvailableFonts objects for all LilyPond installations
# visited at runtime
_available_fonts = {}


def available(lilypond_info):
    """Returns the AvailableFonts object for a given LilyPond version."""
    key = lilypond_info.abscommand() or lilypond_info.command()
    if key not in _available_fonts.keys():
        _available_fonts[key] = AvailableFonts(lilypond_info)
    return _available_fonts[key]


_music_fonts_repo = None


def music_fonts_repo():
    return _music_fonts_repo


def set_music_fonts_repo():
    global _music_fonts_repo
    settings_path = QSettings().value('music-fonts/font-repo', '', str)
    if not settings_path:
        _music_fonts_repo = None
    else:
        if (
            not _music_fonts_repo
            or _music_fonts_repo.root() != settings_path
        ):
            _music_fonts_repo = musicfonts.MusicFontRepo(settings_path)


set_music_fonts_repo()
app.settingsChanged.connect(set_music_fonts_repo)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Paths preferences page
"""


from PyQt6.QtCore import QSettings
from PyQt6.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
)

import app
import preferences
import qsettings
import widgets.listedit
import widgets.urlrequester


class Paths(preferences.GroupsPage):
    def __init__(self, dialog):
        super().__init__(dialog)

        layout = QVBoxLayout()
        self.setLayout(layout)

        layout.addWidget(HyphenPaths(self))
        layout.addWidget(MusicFonts(self))
        layout.addStretch(1)


class HyphenPaths(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.listedit = widgets.listedit.FilePathEdit()
        self.listedit.changed.connect(self.changed)
        layout.addWidget(self.listedit)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Folders containing hyphenation dictionaries"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("hyphenation")
        paths = qsettings.get_string_list(s, "paths")
        self.listedit.setValue(paths)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("hyphenation")
        paths = self.listedit.value()
        if paths:
            s.setValue("paths", paths)
        else:
            s.remove("paths")


class MusicFonts(preferences.Group):
    """Preferences regarding the handling of music/notation fonts."""

    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        font_repo_layout = QHBoxLayout()
        self.font_repo_label = QLabel()
        font_repo_layout.addWidget(self.font_repo_label)
        self.font_repo_cb = QCheckBox()
        self.font_repo_cb.toggled.connect(self.changed)
        font_repo_layout.addWidget(self.font_repo_cb)
        font_repo_layout.addStretch()
        layout.addLayout(font_repo_layout)

        self.font_repo_path_requester = widgets.urlrequester.UrlRequester()
        self.font_repo_path_requester.changed.connect(self.changed)
        layout.addWidget(self.font_repo_path_requester)

        self.font_cache_label = QLabel()
        layout.addWidget(self.font_cache_label)

        self.font_cache_path_requester = widgets.urlrequester.UrlRequester()
        self.font_cache_path_requester.changed.connect(self.changed)
        layout.addWidget(self.font_cache_path_requester)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Music Fonts"))
        self.font_repo_label.setText(_("Music Font Repository:"))
        repo_tt = _(
            "If set this directory can be used to automatically\n"
            + "install all music fonts into a given LilyPond installation."
        )
        self.font_repo_label.setToolTip(repo_tt)
        self.font_repo_path_requester.setToolTip(repo_tt)
        self.font_repo_cb.setText(_("Auto install"))
        self.font_repo_cb.setToolTip(_(
            "Always install fonts from the music font repository\n"
            "to the current LilyPond installation when opening\n"
            "the Document Fonts dialog."
        ))
        self.font_cache_label.setText(_("Music Font Preview Cache:"))
        cache_tt = _(
            "If a writable path is set the provided notation font samples\n"
            + "are cached persistently, otherwise they are cached in the\n"
            + "operating system's temporary directory where they may be\n"
            + "removed upon shutdown.\n"
            + "Notation font previews from custom files or the active\n"
            + "document are always cached temporarily and removed when\n"
            + "closing Frescobaldi."
        )
        self.font_cache_label.setToolTip(cache_tt)
        self.font_cache_path_requester.setToolTip(cache_tt)

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("music-fonts")
        self.font_cache_path_requester.setPath(s.value("font-cache", '', str))
        self.font_repo_path_requester.setPath(s.value("font-repo", '', str))
        self.font_repo_cb.setChecked(s.value("auto-install", True, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("music-fonts")
        font_cache = self.font_cache_path_requester.path()
        if font_cache:
            s.setValue("font-cache", font_cache)
        else:
            s.remove("font-cache")
        font_repo = self.font_repo_path_requester.path()
        print(font_repo)
        if font_repo:
            s.setValue("font-repo", font_repo)
        else:
            s.remove("font-repo")
        s.setValue("auto-install", self.font_repo_cb.isChecked())
